"""
Code to calculate spin-orbit spillage.
Please cite the following paper if you use the code:
https://www.nature.com/articles/s41598-019-45028-y
https://www.nature.com/articles/s41524-020-0319-4
"""

import sys
import numpy as np
from jarvis.io.vasp.outputs import Wavecar  # vaspwfc
import scipy as sp


class Spillage(object):
    def __init__(self, wf_noso="", wf_so=""):
        self.wf_noso = wf_noso
        self.wf_so = wf_so

    def isclose(self, n1, n2, rel_tol=1e-7):

        if abs(n1 - n2) < rel_tol:
            return True
        else:
            return False

    def orth(self, A):
        u, s, vh = np.linalg.svd(A, full_matrices=False)
        M, N = A.shape
        eps = np.finfo(float).eps
        tol = max(M, N) * np.amax(s) * eps
        num = np.sum(s > tol, dtype=int)
        Q = u[:, :num]
        return Q, num

    def overlap_so_spinpol(self):
        noso = Wavecar(filename=self.wf_noso, lsorbit=False)
        so = Wavecar(filename=self.wf_so, lsorbit=True)
        # band gap stuff. not needed per se, just a useful sanity check
        noso_k, noso_bands = noso.readWFBand()
        so_k, so_bands = so.readWFBand()

        # changes section 1
        nelec_list = []
        for nk1 in range(1, noso._nkpts + 1):  # no spin orbit kpoints loop
            knoso = noso._kvecs[nk1 - 1, :]
            for nk2 in range(1, so._nkpts + 1):  # spin orbit
                kso = so._kvecs[nk2 - 1, :]
                if (
                    self.isclose(kso[0], knoso[0])
                    and self.isclose(kso[1], knoso[1])
                    and self.isclose(kso[2], knoso[2])
                ):  # do kpoints match?
                    for c, e in enumerate(noso._occs[0, nk1 - 1, :]):
                        if e < 0.5:
                            cup = c
                            break
                    for c, e in enumerate(noso._occs[1, nk1 - 1, :]):
                        if e < 0.5:
                            cdn = c
                            break

                    nelec_list.append([cup, cdn, cup + cdn])

        n_arr = np.array(nelec_list)

        n_up = int(round(np.mean(n_arr[:, 0])))
        n_dn = int(round(np.mean(n_arr[:, 1])))
        n_tot = int(round(np.mean(n_arr[:, 2])))

        nelec = int(n_tot)

        noso_homo_up = np.max(noso_bands[0, :, n_up - 1])
        noso_lumo_up = np.min(noso_bands[0, :, n_up])

        noso_homo_dn = np.max(noso_bands[1, :, n_dn - 1])
        noso_lumo_dn = np.min(noso_bands[1, :, n_dn])

        so_homo = np.max(so_bands[0, :, nelec - 1])
        so_lumo = np.min(so_bands[0, :, nelec])

        noso_direct_up = np.min(noso_bands[0, :, n_up] - noso_bands[0, :, n_up - 1])
        noso_direct_dn = np.min(noso_bands[1, :, n_dn] - noso_bands[1, :, n_dn - 1])

        so_direct = np.min(so_bands[0, :, nelec] - so_bands[0, :, nelec - 1])

        noso_direct = 1000000.0
        noso_homo = -10000000.0
        noso_lumo = 100000000.0
        for i in range(noso_bands.shape[1]):
            homo_k = max(noso_bands[0, i, n_up - 1], noso_bands[1, i, n_dn - 1])
            lumo_k = min(noso_bands[0, i, n_up], noso_bands[1, i, n_dn])
            noso_direct = min(noso_direct, lumo_k - homo_k)

            noso_homo = max(noso_homo, homo_k)
            noso_lumo = min(noso_lumo, lumo_k)

        gamma_k = []
        kpoints = []
        np.set_printoptions(precision=4)

        x = []
        y = []

        nelec_tot = 0.0
        for nk1 in range(1, noso._nkpts + 1):  # no spin orbit kpoints loop
            knoso = noso._kvecs[nk1 - 1, :]
            for nk2 in range(1, so._nkpts + 1):  # spin orbit
                kso = so._kvecs[nk2 - 1, :]
                if (
                    self.isclose(kso[0], knoso[0])
                    and self.isclose(kso[1], knoso[1])
                    and self.isclose(kso[2], knoso[2])
                ):  # do kpoints match?

                    # changes section 2
                    nelec_up = n_arr[nk1 - 1, 0]
                    nelec_dn = n_arr[nk1 - 1, 1]
                    nelec_tot = n_arr[nk1 - 1, 2]

                    gamma_k.append(nelec_tot)
                    kpoints.append(kso)
                    Mmn = 0.0
                    vnoso = noso.readBandCoeff(ispin=1, ikpt=nk1, iband=1, norm=False)
                    n_noso1 = vnoso.shape[0]
                    vnoso = noso.readBandCoeff(ispin=2, ikpt=nk1, iband=1, norm=False)
                    n_noso2 = vnoso.shape[0]
                    vso = so.readBandCoeff(ispin=1, ikpt=nk2, iband=1, norm=False)
                    n_so = vso.shape[0]
                    vs = min(n_noso1 * 2, n_so)
                    Vnoso = np.zeros((vs, nelec_tot), dtype=complex)
                    Vso = np.zeros((vs, nelec_tot), dtype=complex)

                    # prepare matricies
                    for n1 in range(1, nelec_up + 1):
                        Vnoso[0 : vs // 2, n1 - 1] = noso.readBandCoeff(
                            ispin=1, ikpt=nk1, iband=n1, norm=False
                        )[0 : vs // 2]
                    for n1 in range(1, nelec_dn + 1):
                        Vnoso[vs // 2 : vs, n1 - 1 + nelec_up] = noso.readBandCoeff(
                            ispin=2, ikpt=nk1, iband=n1, norm=False
                        )[0 : vs // 2]

                    for n1 in range(1, nelec_tot + 1):
                        t = so.readBandCoeff(ispin=1, ikpt=nk2, iband=n1, norm=False)
                        Vso[0 : vs // 2, n1 - 1] = t[0 : vs // 2]
                        Vso[vs // 2 : vs, n1 - 1] = t[n_so // 2 : n_so // 2 + vs // 2]
                    Qnoso, num_noso = self.orth(Vnoso)  # make orthonormal basis?

                    Qso, num_so = self.orth(Vso)

                    # evalute main expression

                    a = []
                    for n1 in range(0, nelec_tot):  # noso occupied bands
                        v1 = Qnoso[:, n1]
                        aa = 0.0
                        for n2 in range(0, nelec_tot):  # so occupied bands
                            v2 = Qso[:, n2]

                            t = np.dot(np.conj(v1), v2)
                            Mmn += t * t.conj()
                            aa += (t * t.conj()).real
                        a.append(aa)
                    gamma_k[-1] -= Mmn  # eq 4 in prb 90 125133
                    x.append(kso)
                    y.append(np.real(gamma_k[-1]))
                    if gamma_k[-1] > 0.5:
                        print(
                            "nk1 nk2 kpoint gamma_k ",
                            nk1,
                            nk2,
                            kso,
                            knoso,
                            np.real(gamma_k[-1]),
                            "!!!!!!!!!!",
                        )

        gmax = max(np.real(gamma_k))
        nkmax = np.argmax(np.real(gamma_k))
        kmax = kpoints[nkmax]

        print("------------------------------------")
        print()
        print("                   INDIRECT DIRECT      HOMO/LUMO (eV)")
        print(
            "no spin-orbit gaps",
            "{:+.3f}".format(float(noso_lumo - noso_homo)),
            "{:+.3f}".format(noso_direct),
            "   ",
            [noso_homo, noso_lumo],
        )
        print(
            "spin-orbit gaps   ",
            "{:+.3f}".format(float(so_lumo - so_homo)),
            "{:+.3f}".format(so_direct),
            "   ",
            [so_homo, so_lumo],
        )
        print()
        info = {}
        print("gamma max", np.real(gmax), " at k =  ", kmax)
        info["spillage"] = np.real(gmax)
        info["spillage_k"] = np.real(gamma_k)
        info["kpoints"] = kpoints
        info["kmax"] = kmax
        info["noso_direct"] = noso_direct
        info["so_direct"] = so_direct
        info["x_axis"] = x
        info["y_axis"] = y
        info["so_lumo"] = so_lumo
        info["so_homo"] = so_homo
        info["noso_lumo"] = noso_lumo
        info["noso_homo"] = noso_homo
        return info

"""
if __name__ == "__main__":
    # JVASP-1044

    wf_noso = "/rk2/knc6/Chern3D/JVASP-1067_mp-541837_PBEBO/MAIN-MAGSCFBAND-bulk@JVASP-1067_mp-541837/WAVECAR"  # sys.argv[1]
    wf_so = "/rk2/knc6/Chern3D/JVASP-1067_mp-541837_PBEBO/MAIN-SOCSCFBAND-bulk@JVASP-1067_mp-541837/WAVECAR"  # sys.argv[2]
    spl = Spillage(wf_noso=wf_noso, wf_so=wf_so)
    info = spl.overlap_so_spinpol()
    print(info["spillage"])
"""



"""
Modules for making crystallographic plane surfaces
"""
from jarvis.io.vasp.inputs import Poscar
from jarvis.core.lattice import Lattice
from jarvis.core.atoms import Atoms
import numpy as np
from jarvis.analysis.structure.spacegroup import Spacegroup3D
from numpy.linalg import norm, solve
from numpy import gcd
import sys
#from pymatgen.analysis.wulff import WulffShape
#from pymatgen.core.lattice import Lattice


def ext_gcd(a, b):
    if b == 0:
        return 1, 0
    elif a % b == 0:
        return 0, 1
    else:
        x, y = self.ext_gcd(b, a % b)
        return y, x - y * (a // b)


def wulff_normals(miller_indices=[], surface_energies=[]):
    max_s = min(surface_energies)
    normals = []
    for i, j in zip(miller_indices, surface_energies):
        normal = j * np.linalg.norm(i) / float(max_s)
        normals.append([normal, i])

    return normals


class Surface(object):
    def __init__(
        self,
        atoms=None,
        indices=[0, 0, 1],
        layers=3,
        vacuum=18.0,
        tol=1e-10,
        from_conventional_structure=True,
    ):
        """
        Get surface object of arbitrary atoms object
        """
        self.indices = np.array(indices)
        if from_conventional_structure:
            self.atoms = Spacegroup3D(atoms).conventional_standard_structure
        else:
            self.atoms = atoms
        self.tol = tol
        self.vacuum = vacuum
        self.layers = layers

    def make_surface(self):
        atoms = self.atoms
        h, k, l = self.indices
        h0, k0, l0 = self.indices == 0

        if h0 and k0 or h0 and l0 or k0 and l0:  # if two indices are zero
            if not h0:
                c1, c2, c3 = [(0, 1, 0), (0, 0, 1), (1, 0, 0)]
            if not k0:
                c1, c2, c3 = [(0, 0, 1), (1, 0, 0), (0, 1, 0)]
            if not l0:
                c1, c2, c3 = [(1, 0, 0), (0, 1, 0), (0, 0, 1)]
        else:
            p, q = ext_gcd(k, l)
            a1, a2, a3 = self.atoms.lattice_mat  # .lat_lengths()

            # constants describing the dot product of basis c1 and c2:
            # dot(c1,c2) = k1+i*k2, i in Z
            k1 = np.dot(p * (k * a1 - h * a2) + q * (l * a1 - h * a3), l * a2 - k * a3)
            k2 = np.dot(l * (k * a1 - h * a2) - k * (l * a1 - h * a3), l * a2 - k * a3)

            if abs(k2) > self.tol:
                i = -int(round(k1 / k2))  # i corresponding to the optimal basis
                p, q = p + i * l, q - i * k

            a, b = ext_gcd(p * k + q * l, h)

            c1 = (p * k + q * l, -p * h, -q * h)
            c2 = np.array((0, l, -k)) // abs(gcd(l, k))
            c3 = (b, a * p, a * q)
        # print ('c1c2c3',c1,c2,c3)
        lattice = atoms.lattice_mat  # .lat_lengths()
        basis = np.array([c1, c2, c3])
        scaled = np.linalg.solve(basis.T, np.array(atoms.frac_coords).T).T
        scaled -= np.floor(scaled + self.tol)
        # atoms.frac_coords=scaled
        new_coords = scaled
        tmp_cell = np.dot(basis, lattice)
        M = np.linalg.solve(lattice, tmp_cell)
        # print ('scaled',scaled)
        cart_coords = np.dot(scaled, lattice)
        # print ('cart_coords',cart_coords)
        # print ('M Matric',M)
        new_coords = np.dot(cart_coords, M)
        # print ('new_coords',new_coords)
        # new_cart_coords=np.dot(scaled,tmp_cell)

        new_atoms = Atoms(
            lattice_mat=tmp_cell,
            coords=new_coords,
            elements=atoms.elements,
            cartesian=True,
        )

        surf_atoms = new_atoms.make_supercell([1, 1, self.layers])
        # print("supercell_cart_coords", surf_atoms.frac_coords)

        new_lat = surf_atoms.lattice_mat  # lat_lengths()
        a1 = new_lat[0]
        a2 = new_lat[1]
        a3 = new_lat[2]
        new_lat = np.array(
            [
                a1,
                a2,
                np.cross(a1, a2)
                * np.dot(a3, np.cross(a1, a2))
                / norm(np.cross(a1, a2)) ** 2,
            ]
        )

        a1 = new_lat[0]
        a2 = new_lat[1]
        a3 = new_lat[2]
        # print("a1,a2,a3", new_lat)

        latest_lat = np.array(
            [
                (np.linalg.norm(a1), 0, 0),
                (
                    np.dot(a1, a2) / np.linalg.norm(a1),
                    np.sqrt(
                        np.linalg.norm(a2) ** 2
                        - (np.dot(a1, a2) / np.linalg.norm(a1)) ** 2
                    ),
                    0,
                ),
                (0, 0, np.linalg.norm(a3)),
            ]
        )

        M = np.linalg.solve(new_lat, latest_lat)

        new_cart_coords = surf_atoms.cart_coords  # np.dot(scaled,lattice)

        new_coords = np.dot(new_cart_coords, M)

        new_atoms = Atoms(
            lattice_mat=latest_lat,
            elements=surf_atoms.elements,
            coords=new_coords,
            cartesian=True,
        )

        frac_coords = new_atoms.frac_coords

        frac_coords[:] = frac_coords[:] % 1
        new_atoms = Atoms(
            lattice_mat=latest_lat,
            elements=surf_atoms.elements,
            coords=frac_coords,
            cartesian=False,
        )
        new_lat = new_atoms.lattice_mat
        new_cart_coords = new_atoms.cart_coords
        elements = new_atoms.elements
        new_lat[2][2] = new_lat[2][2] + self.vacuum
        with_vacuum_atoms = Atoms(
            lattice_mat=new_lat,
            elements=elements,
            coords=new_cart_coords,
            cartesian=True,
        )
        # new_atoms.center()
        # print (with_vacuum_atoms)
        return with_vacuum_atoms

"""
if __name__ == "__main__":
    box = [[2.715, 2.715, 0], [0, 2.715, 2.715], [2.715, 0, 2.715]]
    coords = [[0, 0, 0], [0.25, 0.25, 0.25]]
    elements = ["Si", "Si"]
    Si = Atoms(lattice_mat=box, coords=coords, elements=elements)
    Surface(atoms=Si, indices=[1, 1, 1]).make_surface()
    su = [
        0.8582640971273426,
        0.9334963319196496,
        0.9360461382184894,
        0.9419095687284446,
        0.9802042233627004,
        0.9875446840480956,
        1.0120634294466684,
        1.0126231880823566,
        1.0241538763302507,
        1.0315901848682645,
        1.0318271257831195,
        1.0331286888257398,
        1.0344297141291043,
        1.0388709097092674,
        1.040277640596931,
        1.042494119906149,
        1.04453679643896,
        1.0450598648770613,
        1.045076130339553,
        1.0469310544190567,
        1.0491015867538047,
        1.0495494553198788,
        1.0534717916897114,
        1.0535201391639715,
        1.054233162444997,
        1.0579157863887743,
        1.0595676718662346,
        1.0601381085497692,
        1.109580394178689,
    ]

    ml = [
        [0, 0, 1],
        [2, 0, 3],
        [2, 0, 1],
        [1, 0, 1],
        [3, 0, 2],
        [1, 0, 3],
        [3, 1, 1],
        [3, 0, 1],
        [3, 1, 3],
        [3, -1, 1],
        [3, 1, 0],
        [3, 2, 1],
        [3, 3, 1],
        [1, 0, 0],
        [2, 2, 1],
        [3, -1, 3],
        [3, -1, 2],
        [3, 3, 2],
        [3, 2, 2],
        [2, -1, 3],
        [3, 2, 0],
        [3, 2, 3],
        [1, 1, 1],
        [1, 0, 2],
        [3, 1, 2],
        [2, -1, 2],
        [3, -1, 0],
        [2, 2, 3],
        [1, 1, 0],
    ]
    nm = wulff_normals(miller_indices=ml, surface_energies=su)
    print(nm)
    lat = Lattice([[4.05, 0, 0], [0, 4.05, 0], [0, 0, 4.05]])
    pmg_wulff = WulffShape(lat, ml, su)
    print(pmg_wulff.facets)
"""



"""
Modules for making point-defect vacancies
"""
import pprint
import random
from collections import OrderedDict
from jarvis.io.vasp.inputs import Poscar
from jarvis.core.lattice import Lattice
from jarvis.core.atoms import Atoms
import numpy as np
from jarvis.analysis.structure.spacegroup import Spacegroup3D


def rand_select(x=[]):
    uniq = list(set(x))
    info = {}
    for i, ii in enumerate(x):
        info.setdefault(ii, []).append(i)
    selected = {}
    for i, j in info.items():
        chosen = random.choice(j)
        selected.setdefault(i, chosen)
    # print (info)
    # print (selected)
    return selected


class Vacancy(object):
    def __init__(
        self,
        atoms=None,
        defect_structure=None,
        defect_index=None,
        wyckoff_multiplicity=None,
        symbol=None,
    ):
        """
        Get vacancy structures based on Wyckoff positions
        """
        self._atoms = atoms
        self._defect_index = defect_index
        self._defect_structure = defect_structure
        self._wyckoff_multiplicity = wyckoff_multiplicity
        self._symbol = symbol

    def generate_defects(
        self, enforce_c_size=10.0, on_conventional_cell=False, extend=1
    ):
        atoms = self._atoms
        if on_conventional_cell:
            atoms = Spacegroup3D(atoms).conventional_standard_structure
        a = atoms.lattice.lat_lengths()[0]
        b = atoms.lattice.lat_lengths()[1]
        c = atoms.lattice.lat_lengths()[2]
        if enforce_c_size is not None:
            dim1 = int(float(enforce_c_size) / float(a)) + extend
            dim2 = int(float(enforce_c_size) / float(b)) + extend
            dim3 = int(float(enforce_c_size) / float(c)) + extend
            # atoms = atoms.make_supercell([dim1, dim2, dim3])
            supercell_size = [dim1, dim2, dim3]

        element_list = list(set(atoms.elements))
        #print("s_size=", supercell_size, a, b, c, atoms.lattice.lat_lengths())
        #print("s_sizeomg=", atoms.pymatgen_converter().lattice.abc)
        #print(atoms)
        spg = Spacegroup3D(atoms)
        wyckoffs = spg._dataset["wyckoffs"]
        atoms.props = wyckoffs
        supercell = atoms.make_supercell(supercell_size)
        props = rand_select(supercell.props)
        vacs = []
        for i, j in props.items():
            defect_strt = supercell.remove_site_by_index(j)
            vac = Vacancy(
                atoms=supercell,
                defect_structure=defect_strt,
                defect_index=j,
                wyckoff_multiplicity=i,
                symbol=supercell.elements[j],
            )
            vacs.append(vac)
        return vacs

    def to_dict(self):
        d = OrderedDict()
        d["atoms"] = self._atoms
        d["defect_structure"] = self._defect_structure
        d["defect_index"] = self._defect_structure
        d["wyckoff_multiplicity"] = self._wyckoff_multiplicity
        d["symbol"] = self._symbol
        return d

    def __repr__(self, indent=2):
        return pprint.pformat(self.to_dict(), indent=indent)


"""
if __name__ == "__main__":

    box = [[2.715, 2.715, 0], [0, 2.715, 2.715], [2.715, 0, 2.715]]
    coords = [[0, 0, 0], [0.25, 0.25, 0.25]]
    elements = ["Si", "Si"]
    Si = Atoms(lattice_mat=box, coords=coords, elements=elements)
    Si = Poscar.from_file(
        "/rk2/knc6/JARVIS-DFT/2D-bulk/mp-1143_bulk_LDA/MAIN-ELASTIC-bulk@mp_1143/POSCAR"
    ).atoms
    vacs = Vacancy(atoms=Si).generate_defects()
    for i in vacs:
        print(i)
    spg = Spacegroup3D(Si)
    cvn = spg.conventional_standard_structure
    spg = Spacegroup3D(cvn)
    props = spg._dataset["wyckoffs"]
    Si.props = props
    ss = Si.make_supercell([2, 2, 2])
    props = ss.props
    # print (rand_select(props))
"""



import os

vasp_path = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "..",
    "examples",
    "vasp",
    "SiOptb88"
)

cwd = str(os.getcwd())

def test_vasp():
    os.chdir(vasp_path)
    cmd = str('python master.py')
    os.system(cmd)

os.chdir(cwd)    



"""
Modules for running LAMMPS calculations
"""

from jarvis.analysis.structure.spacegroup import (
    Spacegroup3D,
    symmetrically_distinct_miller_indices,
)
from jarvis.core.atoms import Atoms
from jarvis.io.vasp.inputs import Poscar
from jarvis.io.lammps.inputs import LammpsInput, LammpsData
from jarvis.tasks.lammps.templates.templates import GenericInputs
from jarvis.io.lammps.outputs import analyze_log
from jarvis.analysis.defects.vacancy import Vacancy
from jarvis.analysis.defects.surface import Surface

import os
import subprocess
import json
import sys


class JobFactory(object):

    """
    Class for generic LAMMPS calculations
    """

    def __init__(self, name="", pair_style="", pair_coeff="", control_file=""):
        """
        Used in defining a LAMMPS job
        Args:
            pair_style :  LAMMPS pair_style, e.g. "eam/alloy" 
            pair_coeff : path for pair-coefficients file
            control_file :  control-file with units, include modules 
                          for running LAMMPS calculation , see examples
            name : generic name
        """
        self.pair_style = pair_style
        self.pair_coeff = pair_coeff
        self.control_file = control_file
        self.name = name

    def all_props_eam_alloy(
        self,
        atoms=None,
        ff_path="",
        lammps_cmd="",
        enforce_conventional_structure=True,
        enforce_c_size=0,
        extend=1,
    ):
        """
        Generic function for high-throughput LAMMPS calculations using eam/alloy
        Args:
            atoms :  Atoms object
            ff_path :  inter-atomic potential path
            lammps_cmd : LAMMPS executable path
            enforce_conventional_structure :  whether to enforce conventional cell
            enforce_c_size : minimum cell-sizes
            extend : used for round-off during making supercells
        """

        if enforce_conventional_structure:
            atoms = Spacegroup3D(atoms).conventional_standard_structure

        a = atoms.lattice.lat_lengths()[0]
        b = atoms.lattice.lat_lengths()[1]
        c = atoms.lattice.lat_lengths()[2]
        if enforce_c_size is not None:
            dim1 = int(float(enforce_c_size) / float(a)) + extend
            dim2 = int(float(enforce_c_size) / float(b)) + extend
            dim3 = int(float(enforce_c_size) / float(c)) + extend
            atoms = atoms.make_supercell([dim1, dim2, dim3])

        self.pair_style = "eam/alloy"
        self.pair_coeff = ff_path
        parameters = {
            "pair_style": self.pair_style,
            "atom_style": "charge",
            "pair_coeff": self.pair_coeff,
        }
        parameters["control_file"] = "inelast.mod"
        en, final_str, forces = LammpsJob(
            atoms=atoms, jobname="ELASTIC", parameters=parameters, lammps_cmd=lammps_cmd
        ).runjob()
        print("en, final_str, forces", en, final_str, forces)

        indices = symmetrically_distinct_miller_indices(max_index=1, cvn_atoms=atoms)
        for i in indices:
            surf = Surface(atoms=final_str, indices=i).make_surface()
            jobname = str("Surf-") + str("_".join(map(str, i)))
            en2, final_str2, forces2 = LammpsJob(
                atoms=surf,
                jobname=jobname,
                parameters=parameters,
                lammps_cmd=lammps_cmd,
            ).runjob()

        sys.exit()

        v = Vacancy(atoms=final_str).generate_defects(enforce_c_size=5)
        print("vacs=", v)
        for i, ii in enumerate(v):
            jobname = (
                str("symbol-")
                + str(ii._symbol)
                + str("-")
                + str("Wycoff-")
                + str(ii._wyckoff_multiplicity)
            )
            print("ii._defect_structure", ii._atoms)
            en2, final_str2, forces2 = LammpsJob(
                atoms=ii._defect_structure,
                jobname=jobname,
                parameters=parameters,
                lammps_cmd=lammps_cmd,
            ).runjob()

    def optimize_and_elastic(self):
        pass

    def surface_energy(self):
        pass

    def vacancy(self):
        pass

    def phonon(self):
        pass


class LammpsJob(object):
    """
    Class representing a LAMMPS job
    """

    def __init__(
        self,
        atoms=None,
        parameters={
            "pair_style": "eam/alloy",
            "pair_coeff": "abc.alloy",
            "atom_style": "charge",
            "control_file": "inelast.mod",
        },
        lammps_cmd=None,
        output_file="lammps.out",
        stderr_file="std_err.txt",
        jobname="testt",
        attempts=5,
        copy_files=[],
        element_order=[],
    ):
        """
        Used for defining a LAMMPS job
        Args:
            atoms :  Atoms object
            element_order : element order used in accessing force-field parameters
            parameters :  LAMMPS input parameter dictionary
            lammps_cmd : LAMMPS executable path
            output_file :  standard output file
            stderr_file :  standard error file
            jobname :  Job name
            attempts :  number of attempts before crashing the job, TODO
            copy_files : copy certain files before a job
        """

        self.atoms = atoms
        self.element_order = element_order
        self.parameters = parameters
        self.lammps_cmd = lammps_cmd
        self.output_file = output_file
        self.stderr_file = stderr_file
        self.jobname = jobname
        self.attempts = attempts
        self.copy_files = copy_files

    def write_input(self):
        """
        Writes LAMMPS input files
        """
        lmp = LammpsData().atoms_to_lammps(atoms=self.atoms)
        self.element_order = lmp._element_order
        lmp.write_file("data")
        LammpsInput(LammpsDataObj=lmp).write_lammps_in(parameters=self.parameters)
        for i in self.copy_files:
            shutil.copy2(i, ".")
        if "control_file" in self.parameters:
            if self.parameters["control_file"] == "inelast.mod":
                GenericInputs().elastic_general(path=".")

    def run(self):
        """
        Subprocess run a job
        """
        with open(self.output_file, "w") as f_std, open(
            self.stderr_file, "w", buffering=1
        ) as f_err:
            p = subprocess.Popen(
                self.lammps_cmd, shell=True, stdout=f_std, stderr=f_err
            )
            p.wait()
        return p

    def runjob(self):
        """
        Generic LAMMPS job submission
        """
        attempt = 0
        wait = False
        while wait == False:
            attempt = attempt + 1
            if attempt == self.attempts:
                wait = True
            folder = str(os.getcwd()) + str("/") + str(self.jobname)
            if not os.path.exists(folder):
                os.makedirs(folder)
            os.chdir(folder)

            if os.path.isfile("./log.lammps"):
                try:

                    (
                        en,
                        press,
                        toten,
                        c11,
                        c22,
                        c33,
                        c12,
                        c13,
                        c23,
                        c44,
                        c55,
                        c66,
                        c14,
                        c16,
                        c24,
                        c25,
                        c26,
                        c34,
                        c35,
                        c36,
                        c45,
                        c46,
                        c56,
                    ) = analyze_log("./log.lammps")
                    wait = True
                except:
                    pass
                # print ('toten',toten)
            else:
                self.write_input()
                self.run()
                try:

                    (
                        en,
                        press,
                        toten,
                        c11,
                        c22,
                        c33,
                        c12,
                        c13,
                        c23,
                        c44,
                        c55,
                        c66,
                        c14,
                        c16,
                        c24,
                        c25,
                        c26,
                        c34,
                        c35,
                        c36,
                        c45,
                        c46,
                        c56,
                    ) = analyze_log("./log.lammps")
                    wait = True
                except:
                    pass
            pot = os.path.join(os.getcwd(), "potential.mod")
            # print ('toten2',toten,pot)
            initial_str = LammpsData().read_data(
                filename="data", element_order=self.element_order, potential_file=pot
            )
            final_str = LammpsData().read_data(
                potential_file=pot, filename="data0", element_order=self.element_order
            )
            forces = []

            data_cal = []
            data_cal.append(
                {
                    "jobname": self.jobname,
                    "initial_pos": initial_str.to_dict(),
                    "pair_style": str(self.parameters["pair_style"]),
                    "pair_coeff": str(self.parameters["pair_coeff"]),
                    "final_energy": float(toten),
                    "en": en,
                    "press": press,
                    "final_str": final_str.to_dict(),
                }
            )

            json_file = str(self.jobname) + str(".json")
            os.chdir("../")
            f_json = open(json_file, "w")
            f_json.write(json.dumps(data_cal))
            f_json.close()
            return en, final_str, forces


"""
if __name__ == "__main__":
    p = Poscar.from_file(
        "/rk2/knc6/JARVIS-FF/COMB/ffield.comb3.NiAlO_nist/bulk@mp-1143_fold/bulk@mp-1143/new_pymatgen_slab.vasp"
    )
    atoms = Poscar.from_file(
        "/rk2/knc6/JARVIS-FF/FS/Al1.eam.fs_nist/bulk@mp-134_fold/mp-134/new_pymatgen_slab.vasp"
    ).atoms

    cvn = Spacegroup3D(atoms).conventional_standard_structure
    parameters = {
        "pair_style": "eam/alloy",
        "pair_coeff": "/data/knc6/JARVIS-FF-NEW/FS/Al1.eam.fs",
        "atom_style": "charge",
        "control_file": "inelast.mod",
    }

    cmd = "/users/knc6/Software/LAMMPS/lammps-master/src/lmp_serial<in.main"
    # LammpsJob(atoms=cvn, parameters=parameters, lammps_cmd=cmd).runjob()
    # final_str = LammpsData().read_data(
    #            potential_file='/users/knc6/Software/J2020/jarvis/jarvis/tasks/lammps/ELASTIC/potential.mod', filename="/users/knc6/Software/J2020/jarvis/jarvis/tasks/lammps/ELASTIC/data", element_order=[])
    # print ('final_str',final_str)

    cmd = "/users/knc6/Software/LAMMPS/lammps-master/src/lmp_serial<in.main"
    JobFactory().all_props_eam_alloy(
        atoms=cvn, ff_path="/data/knc6/JARVIS-FF-NEW/FS/Al1.eam.fs", lammps_cmd=cmd
    )
"""



"""
Class for analyzing  WT.out file
"""


class WTOut(object):
    def __init__(self, path=""):
        self.path = path

    def get_z2_index(self):
        f = open(self.path, "r")
        lines = f.read().splitlines()
        f.close()
        strng = 0
        for ii, i in enumerate(lines):
            if "# z2 number for 6 planes" in i:
                for j in range(6):
                    tmp = lines[ii + j + 1].split()

                    if (
                        tmp[0] == "k1=0.0,"
                        or tmp[0] == "k2=0.0,"
                        or tmp[0] == "k3=0.0,"
                        or tmp[0] == "k3=0.5,"
                        or tmp[0] == "k2=0.5,"
                        or tmp[0] == "k1=0.5,"
                    ):
                        val = tmp[3]
                        strng = strng + float(val)
                    if tmp[0] == "k1=0.5,":
                        weak1 = tmp[3]
                    if tmp[0] == "k2=0.5,":
                        weak2 = tmp[3]
                    if tmp[0] == "k3=0.5,":
                        weak3 = tmp[3]
        index = (
            str(int(strng % 2))
            + str(";")
            + str(weak1)
            + str(",")
            + str(weak2)
            + str(",")
            + str(weak3)
        )
        return index

    def get_chern_number(self):
        f = open(self.path, "r")
        lines = f.read().splitlines()
        f.close()
        chrn = []
        try:
            for j, jj in enumerate(lines):
                if "# Chern number for 6 planes" in jj:
                    for k in range(1, 7):
                        tmp = float(lines[j + k].split(":")[1])
                        if tmp not in chrn:
                            chrn.append(tmp)
        except:
            pass
        return chrn


"""
if __name__ == "__main__":
    wt = "/rk2/knc6/Chern3D/JVASP-1067_mp-541837_PBEBO/MAIN-WANN-SOC-bulk@JVASP-1067_mp-541837/WT.out"
    z2 = WTOut(path=wt).get_z2_index()
    print(z2)
    chrn = WTOut(path=wt).get_chern_number()
    print(chrn)
"""



"""
Functions to analze LAMMPS output
"""

import sys


def analyze_log(log="log.lammps"):

    """ 
    Analyzes log.lammps file,
    Please note, the output format heavily depends on the input file
    A generic inpu is taken here
    Args:
        log: path to log.lammps file
    Returns:
          en: energy/atom
          press: pressure
          toten: total energy
          cij: elastic constants
    """
    en = 0
    press = 0
    c11 = 0
    c22 = 0
    c33 = 0
    c44 = 0
    c55 = 0
    c66 = 0
    c12 = 0
    c13 = 0
    c23 = 0
    c14 = 0
    c15 = 0
    c16 = 0
    c14 = 0
    c24 = 0
    c25 = 0
    c26 = 0
    c34 = 0
    c35 = 0
    c36 = 0
    c45 = 0
    c46 = 0
    c56 = 0
    try:
        logfile = open(log, "r")
        lines = logfile.read().splitlines()
        for i, line in enumerate(lines):
            if "Loop time of" in line:
                toten = float(lines[i - 1].split()[12])
                press = float(lines[i - 1].split()[2])
                press = float(press) * 0.0001
                en = float(lines[i - 1].split()[12]) / float(lines[i - 1].split()[17])
                break
        logfile.close()
    except:
        pass
    try:
        logfile = open(log, "r")
        lines = logfile.read().splitlines()
        for i, line in enumerate(lines):
            if 'print "Elastic Constant C11all = ${C11all} ${cunits}"' in line:
                c11 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C22all = ${C22all} ${cunits}"' in line:
                c22 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C33all = ${C33all} ${cunits}"' in line:
                c33 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C12all = ${C12all} ${cunits}"' in line:
                c12 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C13all = ${C13all} ${cunits}"' in line:
                c13 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C23all = ${C23all} ${cunits}"' in line:
                c23 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C44all = ${C44all} ${cunits}"' in line:
                c44 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C55all = ${C55all} ${cunits}"' in line:
                c55 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C66all = ${C66all} ${cunits}"' in line:
                c66 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C14all = ${C14all} ${cunits}"' in line:
                c14 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C16all = ${C16all} ${cunits}"' in line:
                c16 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C24all = ${C24all} ${cunits}"' in line:
                c24 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C25all = ${C25all} ${cunits}"' in line:
                c25 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C26all = ${C26all} ${cunits}"' in line:
                c26 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C34all = ${C34all} ${cunits}"' in line:
                c34 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C35all = ${C35all} ${cunits}"' in line:
                c35 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C36all = ${C36all} ${cunits}"' in line:
                c36 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C45all = ${C45all} ${cunits}"' in line:
                c45 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C46all = ${C46all} ${cunits}"' in line:
                c46 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
            if 'print "Elastic Constant C56all = ${C56all} ${cunits}"' in line:
                c56 = ((str((lines[i + 1])).split("=")[1]).split("GPa"))[0]
        logfile.close()
    except:
        pass
    return (
        round(en, 2),
        round(press, 2),
        float(toten),
        round(float(c11), 1),
        round(float(c22), 1),
        round(float(c33), 1),
        round(float(c12), 1),
        round(float(c13), 1),
        round(float(c23), 1),
        round(float(c44), 1),
        round(float(c55), 1),
        round(float(c66), 1),
        round(float(c14), 1),
        round(float(c16), 1),
        round(float(c24), 1),
        round(float(c25), 1),
        round(float(c26), 1),
        round(float(c34), 1),
        round(float(c35), 1),
        round(float(c36), 1),
        round(float(c45), 1),
        round(float(c46), 1),
        round(float(c56), 1),
    )

"""
if __name__ == "__main__":
    lg = "/rk2/knc6/JARVIS-FF/ALLOY8/Mishin_updated-Ni-Al-Co-2013.eam.alloy_nist/bulk@mp-134_fold/bulk@mp-134/log.lammps"
    x = analyze_log(lg)
    print(x)
"""



import os
from jarvis.io.vasp.outputs import Oszicar
from jarvis.io.vasp.inputs import Poscar, Incar
from jarvis.core.kpoints import Kpoints3D as Kpoints


class VaspJobErrorHandlers(object):
    def __init__(self, run_directory, jobid="", all_errors=[]):
        self.run_directory = run_directory
        self.jobid = jobid
        self.all_errors = all_errors

    def check_errors(self, logfile="vasp.out", timeout=18000):
        os.chdir(self.run_directory)

        errors = []
        # print ("going here 12")
        # try:
        #  run=Vasprun("vasprun.xml")
        #  if run.converged_electronic == False:
        #      errors.append("unconverged_electronic")
        #  if run.converged_ionic == False:
        #      errors.append("unconverged")
        # except:
        #    pass
        # f=open("OUTCAR",'r')
        try:
            oszicar = Oszicar("OSZICAR")
            ionic_steps = len(oszicar.ionic_steps)
            with open("OUTCAR") as f:
                for line in f:

                    if "NSW" in line:
                        try:
                            # nbands = int(d[-1].strip())
                            nsw = int(
                                str(line.split("=")[1]).split("number of steps")[0]
                            )
                            if nsw > 1 and ionic_steps == nsw:
                                errors.append("unconverged")
                        except:
                            pass

                    if "NELM" in line:
                        try:
                            # nbands = int(d[-1].strip())
                            # nsw= int(str(line.split("=")[1]).split("number of steps")[0])
                            nelm = int(str(line.split("=")[1]).split(";")[0])
                            electronic_steps = int(oszicar.electronic_steps[-1][1])
                            if electronic_steps < nelm:
                                print("Electronically converged")
                            else:
                                errors.append("unconverged_electronic")

                        except:
                            pass
        except:

            pass

        with open(logfile, "r") as f:
            for line in f:
                l = line.strip()
                if "WARNING: Sub-Space-Matrix is not hermitian in" in line:
                    err = "subspacematrix"
                    if err not in errors:
                        errors.append(err)
                if "Tetrahedron method fails for NKPT<" in line:
                    err = "tet"
                    if err not in errors:
                        errors.append(err)
                if "Fatal error detecting k-mesh" in line:
                    err = "tet"
                    if err not in errors:
                        errors.append(err)
                if "Fatal error: unable to match k-point" in line:
                    err = "tet"
                    if err not in errors:
                        errors.append(err)
                if "Routine TETIRR needs special values" in line:
                    err = "tet"
                    if err not in errors:
                        errors.append(err)
                if "Tetrahedron method fails for NKPT<" in line:
                    err = "tet"
                    if err not in errors:
                        errors.append(err)
                if "inverse of rotation matrix was not found (increase" in line:
                    err = "inv_rot_ma"
                    if err not in errors:
                        errors.append(err)
                if "SYMPREC" in line:
                    err = "inv_rot_ma"
                    if err not in errors:
                        errors.append(err)
                if "Routine TETIRR needs special values" in line:
                    err = "tetirr"
                    if err not in errors:
                        errors.append(err)
                if "Could not get correct shift" in line:
                    err = "incorrect_shift"
                    if err not in errors:
                        errors.append(err)
                if "REAL_OPTLAY: internal error" in line:
                    err = "real_optlay"
                    if err not in errors:
                        errors.append(err)
                if "REAL_OPT: internal ERROR" in line:
                    err = "real_optlay"
                    if err not in errors:
                        errors.append(err)
                if "ERROR RSPHER" in line:
                    err = "rspher"
                    if err not in errors:
                        errors.append(err)
                if "DENTET" in line:
                    err = "dentet"
                    if err not in errors:
                        errors.append(err)
                if "TOO FEW BAND" in line:
                    err = "too_few_bands"
                    if err not in errors:
                        errors.append(err)
                if "ERROR: the triple product of the basis vectors" in line:
                    err = "triple_product"
                    if err not in errors:
                        errors.append(err)
                if "Found some non-integer element in rotation matrix" in line:
                    err = "rot_matrix"
                    if err not in errors:
                        errors.append(err)
                if "BRIONS problems: POTIM should be increased" in line:
                    err = "brions"
                    if err not in errors:
                        errors.append(err)
                if "internal error in subroutine PRICEL" in line:
                    err = "pricel"
                    if err not in errors:
                        errors.append(err)
                if "LAPACK: Routine ZPOTRF failed" in line:
                    err = "zpotrf"
                    if err not in errors:
                        errors.append(err)
                if "One of the lattice vectors is very long (>50 A), but AMIN" in line:
                    err = "amin"
                    if err not in errors:
                        errors.append(err)
                if "ZBRENT: fatal internal in" in line:
                    err = "zbrent"
                    if err not in errors:
                        errors.append(err)
                if "ZBRENT: fatal error in bracketing" in line:
                    err = "zrbent"
                    if err not in errors:
                        errors.append(err)

                if "ERROR in subspace rotation PSSYEVX" in line:
                    err = "pssyevx"
                    if err not in errors:
                        errors.append(err)
                if "WARNING in EDDRMM: call to ZHEGV failed" in line:
                    err = "eddrmm"
                    if err not in errors:
                        errors.append(err)
                if "Error EDDDAV: Call to ZHEGV failed" in line:
                    err = "edddav"
                    if err not in errors:
                        errors.append(err)
                if "Your FFT grids (NGX,NGY,NGZ) are not sufficient" in line:
                    err = "aliasing_incar"
                    if err not in errors:
                        errors.append(err)

        # UNCONVERGED
        st = os.stat(logfile)
        if time.time() - st.st_mtime > timeout:
            errors.append("Frozenjob")

        all_errors = list(set(errors))
        self.all_errors = all_errors
        return all_errors

    def correct_errors(self):
        msg = self.all_errors
        mat = Poscar.from_file("POSCAR")
        incar = Incar.from_file("INCAR")
        kpoints = Kpoints.from_file("KPOINTS")
        while len(msg) > 0:

            if "zpotrf" in msg:
                print("zpotrf error")

                try:
                    oszicar = Oszicar("OSZICAR")
                    nsteps = len(oszicar.ionic_steps)
                except:
                    nsteps = 0

                if nsteps >= 1:
                    potim = float(incar.get("POTIM", 0.5)) / 2.0
                    incar.update({"ISYM": 0, "POTIM": potim})
                    # incar.write_file("INCAR")
                else:
                    s = mat.atoms
                    comm = mat.comment
                    s.apply_strain(0.05)
                    pos = Poscar(s)
                    pos.comment = comm + str("corrected")
                    # line = str(pos.comment) + str(" ") + str(s) + "\n"
                    # correc.write(line)
                    # pos.write_file("POSCAR")
                    try:
                        cmd = str("rm -rf WAVECAR CHGCAR")
                        os.system(cmd)
                    except:
                        pass

            # if "brions" in msg:
            #     potim = float(incar.get("POTIM", 0.5)) +0.1
            #     incar.update({"POTIM":potim})

            if "zbrent" in msg:
                incar.update({"IBRION": 1})
            if "too_few_bands" in msg:
                with open("OUTCAR") as f:
                    for line in f:
                        if "NBANDS" in line:
                            try:
                                d = line.split("=")
                                nbands = int(d[-1].strip())
                                incar.update({"NBANDS": int(1.1 * nbands)})
                                break
                            except (IndexError, ValueError):
                                pass

            if "pssyevx" in msg:
                incar.update({"ALGO": "Normal"})
            if "edddav" in msg:
                incar.update({"LPLANE": "False"})
                incar.update({"ALGO": "Damped"})
                try:
                    cmd = str("rm -rf CHGCAR")
                    os.system(cmd)
                except:
                    pass
            if "pricel" in msg:
                incar.update({"ALGO": "Normal"})
            if "subspacematrix" in msg:
                incar.update({"NPAR": 1})
            if "rspher" in msg:
                incar.update({"NPAR": 1})
            if "real_optlay" in msg:
                incar.update({"NPAR": 1})
            if "subspacematrix" in msg:
                incar.update({"NPAR": 1})

            if "rot_matrix" in msg:
                incar.update({"ISYM": 0})
            if "amin" in msg:
                incar.update({"AMIN": 0.01})
            if "FrozenJob" in msg:
                try:
                    cmd = str("qdel ") + str(self.job_id)
                    os.system(cmd)
                except:
                    pass

                try:
                    cmd = str("scancel ") + str(self.job_id)
                    os.system(cmd)
                except:
                    pass
                # incar.update({"EDIFF":1e-8,"NSW":1,"IBRION":1})
            # Unconverged
            if "unconverged_electronic" in msg:
                nelm = int(incar.get("NELM", 60))
                incar.update({"NELM": int(nelm) + 100})
            incar.write_file("INCAR")
            kpoints.write_file("KPOINTS")
            potcar.write_file("POTCAR")
            mat.write_file("POSCAR")
            if "unconverged" in msg and os.path.isfile("OSZICAR"):
                oszicar = Oszicar("OSZICAR")
                nsteps = len(oszicar.ionic_steps)

                with open("OUTCAR") as f:
                    for line in f:

                        if "NSW" in line:
                            try:
                                nsw = int(
                                    str(line.split("=")[1]).split("number of steps")[0]
                                )
                                incar.update({"EDIFF": 1e-8, "NSW": nsw})
                                shutil.copy2("POSCAR", "POSCAR.orig")
                                # shutil.copy2('CONTCAR','POSCAR')
                            except:
                                pass
                if nsw > 1 and nsteps == nsw:
                    shutil.copy2("CONTCAR", "POSCAR")



from itertools import product
import numpy as np
from jarvis.core.atoms import Atoms
from jarvis.io.vasp.inputs import Poscar
from jarvis.core.lattice import Lattice
import json
import joblib, pickle


class ZSLGenerator(object):
    """
    Uses Zur algorithm to find best matched interfaces
    This class is modified from pymatgen.
    """

    def __init__(
        self,
        max_area_ratio_tol=0.09,
        max_area=400,
        max_length_tol=0.03,
        max_angle_tol=0.01,
    ):
        """
        Intialize a Zur Super Lattice Generator for a specific film and
            substrate
        Args:
            max_area_ratio_tol(float): Max tolerance on ratio of
                super-lattices to consider equal
            max_area(float): max super lattice area to generate in search
            max_length_tol: maximum length tolerance in checking if two
                vectors are of nearly the same length
            max_angle_tol: maximum angle tolerance in checking of two sets
                of vectors have nearly the same angle between them
        """
        self.max_area_ratio_tol = max_area_ratio_tol
        self.max_area = max_area
        self.max_length_tol = max_length_tol
        self.max_angle_tol = max_angle_tol

    def is_same_vectors(self, vec_set1, vec_set2):
        """
        Determine if two sets of vectors are the same within length and angle
        tolerances
        Args:
            vec_set1(array[array]): an array of two vectors
            vec_set2(array[array]): second array of two vectors
        """
        if np.absolute(rel_strain(vec_set1[0], vec_set2[0])) > self.max_length_tol:
            return False
        elif np.absolute(rel_strain(vec_set1[1], vec_set2[1])) > self.max_length_tol:
            return False
        elif np.absolute(rel_angle(vec_set1, vec_set2)) > self.max_angle_tol:
            return False
        else:
            return True

    def generate_sl_transformation_sets(self, film_area, substrate_area):
        """
        Generates transformation sets for film/substrate pair given the
        area of the unit cell area for the film and substrate. The
        transformation sets map the film and substrate unit cells to super
        lattices with a maximum area
        Args:
            film_area(int): the unit cell area for the film
            substrate_area(int): the unit cell area for the substrate
        Returns:
            transformation_sets: a set of transformation_sets defined as:
                1.) the transformation matricies for the film to create a
                super lattice of area i*film area
                2.) the tranformation matricies for the substrate to create
                a super lattice of area j*film area
        """
        transformation_indicies = [
            (i, j)
            for i in range(1, int(self.max_area / film_area))
            for j in range(1, int(self.max_area / substrate_area))
            if np.absolute(film_area / substrate_area - float(j) / i)
            < self.max_area_ratio_tol
        ]

        # Sort sets by the square of the matching area and yield in order
        # from smallest to largest
        for x in sorted(transformation_indicies, key=lambda x: x[0] * x[1]):
            yield (gen_sl_transform_matricies(x[0]), gen_sl_transform_matricies(x[1]))

    def get_equiv_transformations(
        self, transformation_sets, film_vectors, substrate_vectors
    ):
        """
        Applies the transformation_sets to the film and substrate vectors
        to generate super-lattices and checks if they matches.
        Returns all matching vectors sets.
        Args:
            transformation_sets(array): an array of transformation sets:
                each transformation set is an array with the (i,j)
                indicating the area multipes of the film and subtrate it
                corresponds to, an array with all possible transformations
                for the film area multiple i and another array for the
                substrate area multiple j.
            film_vectors(array): film vectors to generate super lattices
            substrate_vectors(array): substrate vectors to generate super
                lattices
        """

        for (film_transformations, substrate_transformations) in transformation_sets:
            # Apply transformations and reduce using Zur reduce methodology
            films = [
                reduce_vectors(*np.dot(f, film_vectors)) for f in film_transformations
            ]

            substrates = [
                reduce_vectors(*np.dot(s, substrate_vectors))
                for s in substrate_transformations
            ]

            # Check if equivalant super lattices
            for (f_trans, s_trans), (f, s) in zip(
                product(film_transformations, substrate_transformations),
                product(films, substrates),
            ):
                if self.is_same_vectors(f, s):
                    yield [f, s, f_trans, s_trans]

    def __call__(self, film_vectors, substrate_vectors, lowest=False):
        """
        Runs the ZSL algorithm to generate all possible matching
        :return:
        """

        film_area = vec_area(*film_vectors)
        substrate_area = vec_area(*substrate_vectors)

        # Generate all super lattice comnbinations for a given set of miller
        # indicies
        transformation_sets = self.generate_sl_transformation_sets(
            film_area, substrate_area
        )

        # Check each super-lattice pair to see if they match
        for match in self.get_equiv_transformations(
            transformation_sets, film_vectors, substrate_vectors
        ):
            # Yield the match area, the miller indicies,
            yield self.match_as_dict(
                match[0],
                match[1],
                film_vectors,
                substrate_vectors,
                vec_area(*match[0]),
                match[2],
                match[3],
            )

            # Just want lowest match per direction
            if lowest:
                break

    def match_as_dict(
        self,
        film_sl_vectors,
        substrate_sl_vectors,
        film_vectors,
        substrate_vectors,
        match_area,
        film_transformation,
        substrate_transformation,
    ):
        """
        Returns dict which contains ZSL match
        Args:
            film_miller(array)
            substrate_miller(array)
        """
        d = {}
        d["film_sl_vecs"] = np.asarray(film_sl_vectors)
        d["sub_sl_vecs"] = np.asarray(substrate_sl_vectors)
        d["match_area"] = match_area
        d["film_vecs"] = np.asarray(film_vectors)
        d["sub_vecs"] = np.asarray(substrate_vectors)
        d["film_transformation"] = np.asarray(film_transformation)
        d["substrate_transformation"] = np.asarray(substrate_transformation)

        return d


def gen_sl_transform_matricies(area_multiple):
    """
    Generates the transformation matricies that convert a set of 2D
    vectors into a super lattice of integer area multiple as proven
    in Cassels:
    Cassels, John William Scott. An introduction to the geometry of
    numbers. Springer Science & Business Media, 2012.
    Args:
        area_multiple(int): integer multiple of unit cell area for super
        lattice area
    Returns:
        matrix_list: transformation matricies to covert unit vectors to
        super lattice vectors
    """
    return [
        np.array(((i, j), (0, area_multiple / i)))
        for i in get_factors(area_multiple)
        for j in range(area_multiple // i)
    ]


def rel_strain(vec1, vec2):
    """
    Calculate relative strain between two vectors
    """
    return fast_norm(vec2) / fast_norm(vec1) - 1


def rel_angle(vec_set1, vec_set2):
    """
    Calculate the relative angle between two vector sets
    Args:
        vec_set1(array[array]): an array of two vectors
        vec_set2(array[array]): second array of two vectors
    """
    return vec_angle(vec_set2[0], vec_set2[1]) / vec_angle(vec_set1[0], vec_set1[1]) - 1


def fast_norm(a):
    """
    Much faster variant of numpy linalg norm
    """
    return np.sqrt(np.dot(a, a))


def vec_angle(a, b):
    """
    Calculate angle between two vectors
    """
    cosang = np.dot(a, b)
    sinang = fast_norm(np.cross(a, b))
    return np.arctan2(sinang, cosang)


def vec_area(a, b):
    """
    Area of lattice plane defined by two vectors
    """
    return fast_norm(np.cross(a, b))


def reduce_vectors(a, b):
    """
    Generate independent and unique basis vectors based on the
    methodology of Zur and McGill
    """
    if np.dot(a, b) < 0:
        return reduce_vectors(a, -b)

    if fast_norm(a) > fast_norm(b):
        return reduce_vectors(b, a)

    if fast_norm(b) > fast_norm(np.add(b, a)):
        return reduce_vectors(a, np.add(b, a))

    if fast_norm(b) > fast_norm(np.subtract(b, a)):
        return reduce_vectors(a, np.subtract(b, a))

    return [a, b]


def get_factors(n):
    """
    Generate all factors of n
    """
    for x in range(1, n + 1):
        if n % x == 0:
            yield x

  
def make_interface(film='',subs='',atol=1,ltol=0.05):
    """
    Returns mismatch and other information as info dict
    """


    z = ZSLGenerator()

    matches = list(z(film.lattice_mat[:2], subs.lattice_mat[:2], lowest=True))

    info = {}
    info["mismatch_u"] = "na"
    info["mismatch_v"] = "na"
    info["mismatch_angle"] = "na"
    info["area1"] = "na"
    info["area2"] = "na"
    info["film_sl"] = "na"#film
    info['matches']= matches
    info["subs_sl"] = "na"#subs

    uv1 = matches[0]["sub_sl_vecs"]
    uv2 = matches[0]["film_sl_vecs"]
    # uv1=[[-8.52917200e+00,  9.12745800e+00,  3.66344517e-17],[1.27937580e+01, 9.12745800e+00, 1.34228735e-15]]
    # uv2=[[7.02403800e+00, 1.05360570e+01, 1.07524571e-15],[-1.40480760e+01,  7.02403800e+00, -4.30098283e-16]]
    u = np.array(uv1)
    v = np.array(uv2)
    a1 = u[0]
    a2 = u[1]
    b1 = v[0]
    b2 = v[1]
    mismatch_u = np.linalg.norm(b1) / np.linalg.norm(a1) - 1
    mismatch_v = np.linalg.norm(b2) / np.linalg.norm(a2) - 1
    angle1 = (
        np.arccos(np.dot(a1, a2) / np.linalg.norm(a1) / np.linalg.norm(a2))
        * 180
        / np.pi
    )
    angle2 = (
        np.arccos(np.dot(b1, b2) / np.linalg.norm(b1) / np.linalg.norm(b2))
        * 180
        / np.pi
    )
    mismatch_angle = abs(angle1 - angle2)
    area1 = np.linalg.norm(np.cross(a1, a2))
    area2 = np.linalg.norm(np.cross(b1, b2))
    uv_substrate = uv1
    uv_film = uv2
    substrate_latt = Lattice(
        np.array([uv_substrate[0][:], uv_substrate[1][:], subs.lattice_mat[2, :]])
    )

    _, __, scell = subs.lattice.find_matches(substrate_latt, ltol=ltol, atol=atol)


    film_latt = Lattice(
        np.array([uv_film[0][:], uv_film[1][:], film.lattice_mat[2, :]])
    )


    scell[2] = np.array([0, 0, 1])
    scell_subs = scell

    _, __, scell = film.lattice.find_matches(film_latt, ltol=ltol, atol=atol)
    scell[2] = np.array([0, 0, 1])
    scell_film = scell


    film = film.make_supercell(scell_film)
    subs = subs.make_supercell(scell_subs)

    lmap = Lattice(
        np.array(
            [subs.lattice_mat[0, :], subs.lattice_mat[1, :], film.lattice_mat[2, :]]
        )
    )

    film = film.center_around_origin()
    subs = subs.center_around_origin()
    info["mismatch_u"] = mismatch_u
    info["mismatch_v"] = mismatch_v
    info["mismatch_angle"] = mismatch_angle
    info["area1"] = area1
    info["area2"] = area2
    info["film_sl"] = film
    info["subs_sl"] = subs

    
    seperation=3
    coords_uniq_sub = np.array(subs.cart_coords)


    coords_uniq_film = np.array(film.cart_coords)

    substrate_top_z = max(np.array(subs.frac_coords)[:, 2])
    substrate_bot_z = min(np.array(subs.frac_coords)[:, 2])

    film_bottom = min(np.array(film.frac_coords)[:, 2])
    film_top = max(np.array(film.frac_coords)[:, 2])

    sub_z = subs.lattice_mat[2, :]
    origin = np.array([0, 0, substrate_top_z])
    shift_normal = sub_z / np.linalg.norm(sub_z) * seperation/ np.linalg.norm(sub_z)

    thickness_sub = abs(substrate_top_z-substrate_bot_z)
    thickness_film = abs(film_top-film_bottom)


    new_coords = []
    lattice_mat = subs.lattice_mat
    elements = []
    for i in subs.frac_coords:
        new_coords.append(i)
    for i in subs.elements:
        elements.append(i)

    for i in film.elements:
        elements.append(i)
    for i in film.frac_coords:
        tmp = i
        tmp[2] = i[2] +shift_normal[2]+(thickness_sub+thickness_film)/2.0


        new_coords.append(tmp)


    interface = Atoms(
        lattice_mat=lattice_mat, elements=elements, coords=new_coords, cartesian=False
    )
    info['interface']=interface
    return info




def mismatch_strts(film=[], subs=[],ltol=0.05, atol=10):
    """
    Returns mismatch and other information as info dict,
    Deprecated, this module will be removes in the next version
    """
    z = ZSLGenerator()
    matches = list(z(film.lattice_mat[:2], subs.lattice_mat[:2], lowest=True))
    info = {}
    info["mismatch_u"] = "na"
    info["mismatch_v"] = "na"
    info["mismatch_angle"] = "na"
    info["area1"] = "na"
    info["area2"] = "na"
    info["film_sl"] = film
    info["subs_sl"] = subs

    uv1 = matches[0]["sub_sl_vecs"]
    uv2 = matches[0]["film_sl_vecs"]
    # uv1=[[-8.52917200e+00,  9.12745800e+00,  3.66344517e-17],[1.27937580e+01, 9.12745800e+00, 1.34228735e-15]]
    # uv2=[[7.02403800e+00, 1.05360570e+01, 1.07524571e-15],[-1.40480760e+01,  7.02403800e+00, -4.30098283e-16]]
    u = np.array(uv1)
    v = np.array(uv2)
    a1 = u[0]
    a2 = u[1]
    b1 = v[0]
    b2 = v[1]
    mismatch_u = np.linalg.norm(b1) / np.linalg.norm(a1) - 1
    mismatch_v = np.linalg.norm(b2) / np.linalg.norm(a2) - 1
    angle1 = (
        np.arccos(np.dot(a1, a2) / np.linalg.norm(a1) / np.linalg.norm(a2))
        * 180
        / np.pi
    )
    angle2 = (
        np.arccos(np.dot(b1, b2) / np.linalg.norm(b1) / np.linalg.norm(b2))
        * 180
        / np.pi
    )
    mismatch_angle = abs(angle1 - angle2)
    area1 = np.linalg.norm(np.cross(a1, a2))
    area2 = np.linalg.norm(np.cross(b1, b2))
    uv_substrate = uv1
    uv_mat2d = uv2
    substrate_latt = Lattice(
        np.array([uv_substrate[0][:], uv_substrate[1][:], subs.lattice_mat[2, :]])
    )
    # to avoid numerical issues with find_mapping
    mat2d_fake_c = film.lattice_mat[2, :] / np.linalg.norm(film.lattice_mat[2, :]) * 5.0
    mat2d_latt = Lattice(np.array([uv_mat2d[0][:], uv_mat2d[1][:], mat2d_fake_c]))
    mat2d_latt_fake = Lattice(
        np.array([film.lattice_mat[0, :], film.lattice_mat[1, :], mat2d_fake_c])
    )
    _, __, scell = subs.lattice.find_matches(substrate_latt, ltol=ltol, atol=atol)
    scell[2] = np.array([0, 0, 1])
    # print("subs lattice", subs.lattice_mat)
    # print("scell", scell)
    subs = subs.make_supercell_matrix(scell)
    _, __, scell = mat2d_latt_fake.find_matches(mat2d_latt, ltol=0.05, atol=1)
    scell[2] = np.array([0, 0, 1])
    film = film.make_supercell_matrix(scell)
    # modify the substrate lattice so that the 2d material can be
    # grafted on top of it
    #lmap = Lattice(
    #    np.array(
    #        [subs.lattice_mat[0, :], subs.lattice_mat[1, :], film.lattice_mat[2, :]]
    #    )
    #)
    #film.lattice = lmap

    lmap = Lattice(
        np.array(
            [subs.lattice_mat[0, :], subs.lattice_mat[1, :], film.lattice_mat[2, :]]
        )
    )
    film.lattice = lmap



    # film.modify_lattice(lmap)
    # print ("film",film)
    # print ("subs",subs)

    info["mismatch_u"] = mismatch_u
    info["mismatch_v"] = mismatch_v
    info["mismatch_angle"] = mismatch_angle
    info["area1"] = area1
    info["area2"] = area2
    info["film_sl"] = film
    info["subs_sl"] = subs
    return info


def get_hetero_type(A={}, B={}):
    stack = "na"
    int_type = "na"
    try:
        # if A['phi']>B['phi']:
        if A["scf_vbm"] - A["avg_max"] < B["scf_vbm"] - B["avg_max"]:
            stack = "BA"
        else:
            C = A
            D = B
            A = D
            B = C
            stack = "AB"
            # tmp=B
            # B=A
            # A=tmp
        vbm_a = A["scf_vbm"] - A["avg_max"]
        vbm_b = B["scf_vbm"] - B["avg_max"]
        cbm_a = A["scf_cbm"] - A["avg_max"]
        cbm_b = B["scf_cbm"] - B["avg_max"]
        #  print ('vbm_a,vbm_b,cbm_b,cbm_a',vbm_a,vbm_b,cbm_b,cbm_a)
        if vbm_a < vbm_b and vbm_b < cbm_b and cbm_b < cbm_a:
            int_type = "I"
        elif vbm_a < vbm_b and vbm_b < cbm_a and cbm_a < cbm_b:
            int_type = "II"
        elif vbm_a < cbm_a and cbm_a < vbm_b and vbm_b < cbm_b:
            int_type = "III"
    except:
        pass
    return int_type, stack


def get_hetero(film, substrate, seperation=3.0):
    # unique site coordinates in the substrate top layers
    coords_uniq_sub = np.array(substrate.cart_coords)

    # unique site coordinates in the 2D material bottom layers
    coords_uniq_film = np.array(film.cart_coords)
   
    substrate_top_z = max(np.array(substrate.cart_coords)[:, 2])
    substrate_bot_z = min(np.array(substrate.cart_coords)[:, 2])
    #print ('substrate_top_z',substrate_top_z- substrate_bot_z)
    film_bottom = min(np.array(film.cart_coords)[:, 2])
    film_top = max(np.array(film.cart_coords)[:, 2])
    #print ('film_bottom',film_top-film_bottom)
    # shift normal to the surface by 'seperation'
    sub_z = substrate.lattice_mat[2, :]
    origin = np.array([0, 0, substrate_top_z])
    shift_normal = sub_z / np.linalg.norm(sub_z) * seperation
    #print ('origin', origin)
    #print ('shift_normal', shift_normal)
    thickness_sub = abs(substrate_top_z-substrate_bot_z)
    thickness_film = abs(film_top-film_bottom)
    #shift_normal =  seperation
    #shift_normal = sub_z / np.linalg.norm(sub_z) * seperation
    # generate all possible interfaces, one for each combination of
    # unique substrate and unique 2d materials site in the layers .i.e
    # an interface structure for each parallel shift
    # interface = 2D material + substrate

    new_coords = []
    lattice_mat = substrate.lattice_mat
    elements = []
    for i in substrate.cart_coords:
        new_coords.append(i)
    for i in substrate.elements:
        elements.append(i)

    for i in film.elements:
        elements.append(i)
    for i in film.cart_coords:
        tmp = i
        #tmp[2] = i[2] + (thickness_sub+thickness_film)
        #tmp[2] = i[2] +(thickness_sub+thickness_film)+seperation
        #tmp = tmp  + shift_normal
        tmp = tmp + origin + shift_normal
        #new_coords.append(i)
        new_coords.append(tmp)

    interface = Atoms(
        lattice_mat=lattice_mat, elements=elements, coords=new_coords, cartesian=True
    )

    return interface


"""
if __name__ == "__main__":
    s1 = Poscar.from_file(
        "/rk2/knc6/JARVIS-DFT/2D-1L/POSCAR-mp-1821-1L.vasp_PBEBO/MAIN-RELAX-Surf-mp-1821/POSCAR"
    )
    s2 = Poscar.from_file(
        "/rk2/knc6/JARVIS-DFT/2D-1L/POSCAR-mp-2815-1L.vasp_PBEBO/MAIN-RELAX-Surf-mp-2815/POSCAR"
    )
    print(s1)
    print(s2)
    info = mismatch_strts(film=s1.atoms, subs=s2.atoms)
    # print (info)
    print((s1.__dict__["atoms"].lattice_mat))
    d = s1.__dict__["atoms"]
    f = open("test.json", "wb")
    joblib.dump(s1, f)
    f.close()

    print(get_hetero(s1.atoms, s2.atoms))
    ff = open("test.json", "rb")
    dd = joblib.load(ff)
    ff.close()
    # print (dd.atoms.lattice_mat)
"""



import os
import json
import datetime

from hashlib import sha1


class MontyEncoder(json.JSONEncoder):
    """
    A Json Encoder which supports the MSONable API, plus adds support for
    numpy arrays, datetime objects, bson ObjectIds (requires bson).
    Usage::
        # Add it as a *cls* keyword when using json.dump
        json.dumps(object, cls=MontyEncoder)
    """

    def default(self, o) -> dict:  # pylint: disable=E0202
        """
        Overriding default method for JSON encoding. This method does two
        things: (a) If an object has a to_dict property, return the to_dict
        output. (b) If the @module and @class keys are not in the to_dict,
        add them to the output automatically. If the object has no to_dict
        property, the default Python json encoder default method is called.
        Args:
            o: Python object.
        Return:
            Python dict representation.
        """
        if isinstance(o, datetime.datetime):
            return {"@module": "datetime", "@class": "datetime", "string": o.__str__()}
        if np is not None:
            if isinstance(o, np.ndarray):
                return {
                    "@module": "numpy",
                    "@class": "array",
                    "dtype": o.dtype.__str__(),
                    "data": o.tolist(),
                }
            if isinstance(o, np.generic):
                return o.item()
        if bson is not None:
            if isinstance(o, bson.objectid.ObjectId):
                return {"@module": "bson.objectid", "@class": "ObjectId", "oid": str(o)}

        try:
            d = o.as_dict()
            if "@module" not in d:
                d["@module"] = u"{}".format(o.__class__.__module__)
            if "@class" not in d:
                d["@class"] = u"{}".format(o.__class__.__name__)
            if "@version" not in d:
                try:
                    parent_module = o.__class__.__module__.split(".")[0]
                    module_version = import_module(
                        parent_module
                    ).__version__  # type: ignore
                    d["@version"] = u"{}".format(module_version)
                except (AttributeError, ImportError):
                    d["@version"] = None
            return d
        except AttributeError:
            return json.JSONEncoder.default(self, o)


class MontyDecoder(json.JSONDecoder):
    """
    A Json Decoder which supports the MSONable API. By default, the
    decoder attempts to find a module and name associated with a dict. If
    found, the decoder will generate a Pymatgen as a priority.  If that fails,
    the original decoded dictionary from the string is returned. Note that
    nested lists and dicts containing pymatgen object will be decoded correctly
    as well.
    Usage:
        # Add it as a *cls* keyword when using json.load
        json.loads(json_string, cls=MontyDecoder)
    """

    def process_decoded(self, d):
        """
        Recursive method to support decoding dicts and lists containing
        pymatgen objects.
        """
        if isinstance(d, dict):
            if "@module" in d and "@class" in d:
                modname = d["@module"]
                classname = d["@class"]
                if classname in MSONable.REDIRECT.get(modname, {}):
                    modname = MSONable.REDIRECT[modname][classname]["@module"]
                    classname = MSONable.REDIRECT[modname][classname]["@class"]
            else:
                modname = None
                classname = None
            if modname and modname not in ["bson.objectid", "numpy"]:
                if modname == "datetime" and classname == "datetime":
                    try:
                        dt = datetime.datetime.strptime(
                            d["string"], "%Y-%m-%d %H:%M:%S.%f"
                        )
                    except ValueError:
                        dt = datetime.datetime.strptime(
                            d["string"], "%Y-%m-%d %H:%M:%S"
                        )
                    return dt

                mod = __import__(modname, globals(), locals(), [classname], 0)
                if hasattr(mod, classname):
                    cls_ = getattr(mod, classname)
                    data = {k: v for k, v in d.items() if not k.startswith("@")}
                    if hasattr(cls_, "from_dict"):
                        return cls_.from_dict(data)
            elif np is not None and modname == "numpy" and classname == "array":
                return np.array(d["data"], dtype=d["dtype"])

            elif (
                (bson is not None)
                and modname == "bson.objectid"
                and classname == "ObjectId"
            ):
                return bson.objectid.ObjectId(d["oid"])

            return {
                self.process_decoded(k): self.process_decoded(v) for k, v in d.items()
            }

        if isinstance(d, list):
            return [self.process_decoded(x) for x in d]

        return d

    def decode(self, s):
        """
        Overrides decode from JSONDecoder.
        :param s: string
        :return: Object.
        """
        d = json.JSONDecoder.decode(self, s)
        return self.process_decoded(d)



"""
Modules for calculating theoretical solar-cell efficiency
See: https://pubs.acs.org/doi/abs/10.1021/acs.chemmater.9b02166
"""
import numpy as np
import os
from jarvis.io.vasp.outputs import Vasprun
from scipy.interpolate import interp1d
from numpy import interp
import scipy.constants as constants
from scipy.integrate import simps
from scipy.constants import physical_constants, speed_of_light
import matplotlib.pyplot as plt

#plt.switch_backend("agg")


class SolarEfficiency(object):
    def __init__(self, formalism="slme"):
        """
        Calculates theoretical solar cell efficiency using SLME or SQ approach
        See https://pubs.acs.org/doi/10.1021/acs.chemmater.9b02166
        """
        self.formalism = formalism

    def calculate_SQ(
        self, bandgap_ev, temperature=300, fr=1, plot_current_voltage=False
    ):
        """
            Args:
                bandgap_ev: bandga in electron-volt
                temperature: temperature in K
            Returns:
                 
            """

        # Defining constants for tidy equations
        c = constants.c  # speed of light, m/s
        h = constants.h  # Planck's constant J*s (W)
        h_e = constants.h / constants.e  # Planck's constant eV*s
        k = constants.k  # Boltzmann's constant J/K
        k_e = constants.k / constants.e  # Boltzmann's constant eV/K
        e = constants.e  # Coulomb

        # Load the Air Mass 1.5 Global tilt solar spectrum
        solar_spectrum_data_file = str(
            os.path.join(os.path.dirname(__file__), "am1.5G.dat")
        )
        solar_spectra_wavelength, solar_spectra_irradiance = np.loadtxt(
            solar_spectrum_data_file, usecols=[0, 1], unpack=True, skiprows=2
        )

        solar_spectra_wavelength_meters = solar_spectra_wavelength * 1e-9

        # need to convert solar irradiance from Power/m**2(nm) into
        # photon#/s*m**2(nm) power is Watt, which is Joule / s
        # E = hc/wavelength
        # at each wavelength, Power * (wavelength(m)/(h(Js)*c(m/s))) = ph#/s
        solar_spectra_photon_flux = solar_spectra_irradiance * (
            solar_spectra_wavelength_meters / (h * c)
        )

        ### Calculation of total solar power incoming
        power_in = simps(solar_spectra_irradiance, solar_spectra_wavelength)

        # calculation of blackbody irradiance spectra
        # units of W/(m**3), different than solar_spectra_irradiance!!! (This
        # is intentional, it is for convenience)
        blackbody_irradiance = (
            2.0 * h * c ** 2 / (solar_spectra_wavelength_meters ** 5)
        ) * (
            1.0
            / (
                (np.exp(h * c / (solar_spectra_wavelength_meters * k * temperature)))
                - 1.0
            )
        )

        # I've removed a pi in the equation above - Marnik Bercx

        # now to convert the irradiance from Power/m**2(m) into photon#/s*m**2(m)
        blackbody_photon_flux = blackbody_irradiance * (
            solar_spectra_wavelength_meters / (h * c)
        )

        # absorbance interpolation onto each solar spectrum wavelength

        # Get the bandgap in wavelength in meters
        bandgap_wavelength = h_e * c / bandgap_ev

        # Only take the part of the wavelength-dependent solar spectrum and
        # blackbody spectrum below the bandgap wavelength
        bandgap_index = np.searchsorted(
            solar_spectra_wavelength_meters, bandgap_wavelength
        )

        bandgap_irradiance = interp(
            np.array([bandgap_wavelength]),
            solar_spectra_wavelength_meters,
            solar_spectra_photon_flux,
        )

        bandgap_blackbody = (
            (2.0 * h * c ** 2 / (bandgap_wavelength ** 5))
            * (1.0 / ((np.exp(h * c / (bandgap_wavelength * k * temperature))) - 1.0))
            * (bandgap_wavelength / (h * c))
        )

        integration_wavelength = np.concatenate(
            (
                solar_spectra_wavelength_meters[:bandgap_index],
                np.array([bandgap_wavelength]),
            ),
            axis=0,
        )

        integration_solar_flux = np.concatenate(
            (solar_spectra_photon_flux[:bandgap_index], bandgap_irradiance), axis=0
        )

        integration_blackbody = np.concatenate(
            (blackbody_photon_flux[:bandgap_index], np.array([bandgap_blackbody])),
            axis=0,
        )

        #  Numerically integrating irradiance over wavelength array
        # Note: elementary charge, not math e!  ## units of A/m**2   W/(V*m**2)
        J_0_r = e * np.pi * simps(integration_blackbody, integration_wavelength)

        J_0 = J_0_r / fr

        #  Numerically integrating irradiance over wavelength array
        # elementary charge, not math e!  ### units of A/m**2   W/(V*m**2)
        J_sc = e * simps(integration_solar_flux * 1e9, integration_wavelength)

        #    J[i] = J_sc - J_0*(1 - exp( e*V[i]/(k*T) ) )
        #   #This formula from the original paper has a typo!!
        #    J[i] = J_sc - J_0*(exp( e*V[i]/(k*T) ) - 1)
        #   #Bercx chapter and papers have the correct formula (well,
        #   the correction on one paper)
        def J(V):
            J = J_sc - J_0 * (np.exp(e * V / (k * temperature)) - 1.0)
            return J

        def power(V):
            p = J(V) * V
            return p

        # A more primitive, but perfectly robust way of getting a reasonable
        # estimate for the maximum power.
        test_voltage = 0
        voltage_step = 0.001
        while power(test_voltage + voltage_step) > power(test_voltage):
            test_voltage += voltage_step

        max_power = power(test_voltage)

        # Calculate the maximized efficience
        efficiency = max_power / power_in

        # This segment isn't needed for functionality at all, but can display a
        # plot showing how the maximization of power by choosing the optimal
        # voltage value works
        if plot_current_voltage:
            V = np.linspace(0, 2, 200)
            plt.plot(V, J(V))
            plt.plot(V, power(V), linestyle="--")
            plt.show()
            print(max_power)

        return efficiency

    def slme(
        self,
        material_energy_for_absorbance_data,
        material_absorbance_data,
        material_direct_allowed_gap,
        material_indirect_gap,
        thickness=50e-6,
        temperature=293.15,
        absorbance_in_inverse_centimeters=False,
        cut_off_absorbance_below_direct_allowed_gap=True,
        plot_current_voltage=False,
    ):
        """
            Calculate the
            IMPORTANT NOTES:
            1) Material calculated absorbance is assumed to be in m^-1, not cm^-1!
                (Most sources will provide absorbance in cm^-1, so be careful.)
            2) The default is to remove absorbance below the direct allowed gap.
                This is for dealing with broadening applied in DFT absorbance
                calculations. Probably not desired for experimental data.
            3) We can calculate at different temperatures if we want to, but 25 C /
                293.15 K is the standard temperature assumed if not specified
            4) If absorbance is in cm^-1, multiply values by 100 to match units
                assumed in code
            Args:
                material_energy_for_absorbance_data:
                material_absorbance_data:
                material_direct_allowed_gap:
                material_indirect_gap:
                thickness:
                temperature:
                absorbance_in_inverse_centimeters:
                cut_off_absorbance_below_direct_allowed_gap:
                plot_current_voltage:
            Returns:
                The calculated maximum efficiency.
            """

        # Defining constants for tidy equations
        c = constants.c  # speed of light, m/s
        h = constants.h  # Planck's constant J*s (W)
        h_e = constants.h / constants.e  # Planck's constant eV*s
        k = constants.k  # Boltzmann's constant J/K
        k_e = constants.k / constants.e  # Boltzmann's constant eV/K
        e = constants.e  # Coulomb

        # Make sure the absorption coefficient has the right units (m^{-1})
        if absorbance_in_inverse_centimeters:
            material_absorbance_data = material_absorbance_data * 100

        # Load the Air Mass 1.5 Global tilt solar spectrum
        solar_spectrum_data_file = str(
            os.path.join(os.path.dirname(__file__), "am1.5G.dat")
        )

        solar_spectra_wavelength, solar_spectra_irradiance = np.loadtxt(
            solar_spectrum_data_file, usecols=[0, 1], unpack=True, skiprows=2
        )

        solar_spectra_wavelength_meters = solar_spectra_wavelength * 1e-9

        delta = material_direct_allowed_gap - material_indirect_gap
        fr = np.exp(-delta / (k_e * temperature))

        # need to convert solar irradiance from Power/m**2(nm) into
        # photon#/s*m**2(nm) power is Watt, which is Joule / s
        # E = hc/wavelength
        # at each wavelength, Power * (wavelength(m)/(h(Js)*c(m/s))) = ph#/s
        solar_spectra_photon_flux = solar_spectra_irradiance * (
            solar_spectra_wavelength_meters / (h * c)
        )

        # Calculation of total solar power incoming
        power_in = simps(solar_spectra_irradiance, solar_spectra_wavelength)

        # calculation of blackbody irradiance spectra
        # units of W/(m**3), different than solar_spectra_irradiance!!! (This
        # is intentional, it is for convenience)
        blackbody_irradiance = (
            2.0 * h * c ** 2 / (solar_spectra_wavelength_meters ** 5)
        ) * (
            1.0
            / (
                (np.exp(h * c / (solar_spectra_wavelength_meters * k * temperature)))
                - 1.0
            )
        )

        # I've removed a pi in the equation above - Marnik Bercx

        # now to convert the irradiance from Power/m**2(m) into photon#/s*m**2(m)
        blackbody_photon_flux = blackbody_irradiance * (
            solar_spectra_wavelength_meters / (h * c)
        )

        # units of nm
        material_wavelength_for_absorbance_data = (
            (c * h_e) / (material_energy_for_absorbance_data + 0.00000001)
        ) * 10 ** 9

        # absorbance interpolation onto each solar spectrum wavelength

        # creates cubic spline interpolating function, set up to use end values
        #  as the guesses if leaving the region where data exists
        material_absorbance_data_function = interp1d(
            material_wavelength_for_absorbance_data,
            material_absorbance_data,
            kind="cubic",
            fill_value=(material_absorbance_data[0], material_absorbance_data[-1]),
            bounds_error=False,
        )

        material_interpolated_absorbance = np.zeros(
            len(solar_spectra_wavelength_meters)
        )
        for i in range(0, len(solar_spectra_wavelength_meters)):
            ## Cutting off absorption data below the gap. This is done to deal
            # with VASPs broadening of the calculated absorption data

            if (
                solar_spectra_wavelength[i]
                < 1e9 * ((c * h_e) / material_direct_allowed_gap)
                or cut_off_absorbance_below_direct_allowed_gap == False
            ):
                material_interpolated_absorbance[i] = material_absorbance_data_function(
                    solar_spectra_wavelength[i]
                )

        absorbed_by_wavelength = 1.0 - np.exp(
            -2.0 * material_interpolated_absorbance * thickness
        )

        #  Numerically integrating irradiance over wavelength array
        # Note: elementary charge, not math e!  ## units of A/m**2   W/(V*m**2)
        J_0_r = (
            e
            * np.pi
            * simps(
                blackbody_photon_flux * absorbed_by_wavelength,
                solar_spectra_wavelength_meters,
            )
        )

        J_0 = J_0_r / fr

        #  Numerically integrating irradiance over wavelength array
        # elementary charge, not math e!  ### units of A/m**2   W/(V*m**2)
        J_sc = e * simps(
            solar_spectra_photon_flux * absorbed_by_wavelength, solar_spectra_wavelength
        )

        #    J[i] = J_sc - J_0*(1 - exp( e*V[i]/(k*T) ) )
        #   #This formula from the original paper has a typo!!
        #    J[i] = J_sc - J_0*(exp( e*V[i]/(k*T) ) - 1)
        #   #Bercx chapter and papers have the correct formula (well,
        #   the correction on one paper)
        def J(V):
            J = J_sc - J_0 * (np.exp(e * V / (k * temperature)) - 1.0)
            return J

        def power(V):
            p = J(V) * V
            return p

        # A more primitive, but perfectly robust way of getting a reasonable
        # estimate for the maximum power.
        test_voltage = 0
        voltage_step = 0.001
        while power(test_voltage + voltage_step) > power(test_voltage):
            test_voltage += voltage_step

        max_power = power(test_voltage)

        # Calculate the maximized efficience
        efficiency = max_power / power_in

        # This segment isn't needed for functionality at all, but can display a
        # plot showing how the maximization of power by choosing the optimal
        # voltage value works
        if plot_current_voltage:
            V = np.linspace(0, 2, 200)
            plt.plot(V, J(V))
            plt.plot(V, power(V), linestyle="--")
            plt.savefig("pp.png")
            plt.close()
            # print(power(V_Pmax))

        return efficiency

"""
if __name__ == "__main__":

    v = Vasprun(
        "/rk2/knc6/JARVIS-DFT/Elements-bulkk/mp-149_bulk_PBEBO/MAIN-MBJ-bulk@mp-149/vasprun.xml"
    )
    dirgap = v.get_dir_gap
    indirgap = v.get_indir_gap
    en, abz = v.avg_absorption_coefficient
    abz = abz * 100
    eff = SolarEfficiency().slme(
        en, abz, indirgap, indirgap, plot_current_voltage=False
    )
    print("SLME", 100 * eff)
    eff = SolarEfficiency().calculate_SQ(indirgap)
    print("SQ", 100 * eff)
"""



"""
Modules handling input files for VASP calculations
"""
from collections import Counter
import pprint
import json
import os
import numpy as np
from jarvis.core.atoms import Atoms
from collections import OrderedDict
from jarvis.core.kpoints import generate_kgrid, Kpoints3D


class Poscar(object):
    """
    Class defining Poscar object 
    Args:
        atoms : Atoms object
        comment : Header of Poscar file
    """

    def __init__(self, atoms, comment="System"):
        self.atoms = atoms
        self.comment = comment

    @staticmethod
    def from_file(filename="POSCAR"):
        """
        Read simple POSCAR file from the path
        """
        with open(filename, "r") as f:
            return Poscar.from_string(f.read())

    def write_file(self, filename):
        f = open(filename, "w")
        header = (
            str(self.comment)
            + str("\n1.0\n")
            + str(self.atoms.lattice_mat[0][0])
            + " "
            + str(self.atoms.lattice_mat[0][1])
            + " "
            + str(self.atoms.lattice_mat[0][2])
            + "\n"
            + str(self.atoms.lattice_mat[1][0])
            + " "
            + str(self.atoms.lattice_mat[1][1])
            + " "
            + str(self.atoms.lattice_mat[1][2])
            + "\n"
            + str(self.atoms.lattice_mat[2][0])
            + " "
            + str(self.atoms.lattice_mat[2][1])
            + " "
            + str(self.atoms.lattice_mat[2][2])
            + "\n"
        )
        order = np.argsort(self.atoms.elements)
        coords = self.atoms.frac_coords
        coords_ordered = np.array(coords)[order]
        elements_ordered = np.array(self.atoms.elements)[order]
        props_ordered = np.array(self.atoms.props)[order]
        check_selective_dynamics = False
        if "T" in "".join(map(str, self.atoms.props[0])):
            middle = (
                " ".join(map(str, Counter(elements_ordered).keys()))
                + "\n"
                + " ".join(map(str, Counter(elements_ordered).values()))
                + "\nSelective dynamics\n"
                + "Direct\n"
            )
        else:
            middle = (
                " ".join(map(str, Counter(elements_ordered).keys()))
                + "\n"
                + " ".join(map(str, Counter(elements_ordered).values()))
                + "\ndirect\n"
            )
        rest = ""
        # print ('repr',self.frac_coords, self.frac_coords.shape)
        for ii, i in enumerate(coords_ordered):
            rest = rest + " ".join(map(str, i)) + " " + str(props_ordered[ii]) + "\n"

        result = header + middle + rest

        f.write(result)
        f.close()

    @staticmethod
    def from_string(lines):
        text = lines.splitlines()
        comment = text[0]
        scale = float(text[1])
        lattice_mat = []
        lattice_mat.append([float(i) for i in text[2].split()])
        lattice_mat.append([float(i) for i in text[3].split()])
        lattice_mat.append([float(i) for i in text[4].split()])
        lattice_mat = np.array(lattice_mat)

        uniq_elements = text[5].split()
        element_count = np.array([int(i) for i in text[6].split()])
        elements = []
        for i, ii in enumerate(element_count):
            for j in range(ii):
                elements.append(uniq_elements[i])
        cartesian = True
        if "d" in text[7] or "D" in text[7]:
            cartesian = False
        # print ('cartesian poscar=',cartesian,text[7])
        num_atoms = int(np.sum(element_count))
        coords = []
        for i in range(num_atoms):
            coords.append([float(i) for i in text[8 + i].split()[0:3]])
        coords = np.array(coords)
        atoms = Atoms(
            lattice_mat=lattice_mat,
            coords=coords,
            elements=elements,
            cartesian=cartesian,
        )
        # print (atoms)
        formula = atoms.composition.formula

        return Poscar(atoms, comment=comment)

    def __repr__(self):
        header = (
            str(self.comment)
            + str("\n1.0\n")
            + str(self.atoms.lattice_mat[0][0])
            + " "
            + str(self.atoms.lattice_mat[0][1])
            + " "
            + str(self.atoms.lattice_mat[0][2])
            + "\n"
            + str(self.atoms.lattice_mat[1][0])
            + " "
            + str(self.atoms.lattice_mat[1][1])
            + " "
            + str(self.atoms.lattice_mat[1][2])
            + "\n"
            + str(self.atoms.lattice_mat[2][0])
            + " "
            + str(self.atoms.lattice_mat[2][1])
            + " "
            + str(self.atoms.lattice_mat[2][2])
            + "\n"
        )
        order = np.argsort(self.atoms.elements)
        coords = self.atoms.frac_coords
        coords_ordered = np.array(coords)[order]
        elements_ordered = np.array(self.atoms.elements)[order]
        props_ordered = np.array(self.atoms.props)[order]
        check_selective_dynamics = False
        if "T" in "".join(map(str, self.atoms.props[0])):
            middle = (
                " ".join(map(str, Counter(elements_ordered).keys()))
                + "\n"
                + " ".join(map(str, Counter(elements_ordered).values()))
                + "\nSelective dynamics\n"
                + "Direct\n"
            )
        else:
            middle = (
                " ".join(map(str, Counter(elements_ordered).keys()))
                + "\n"
                + " ".join(map(str, Counter(elements_ordered).values()))
                + "\ndirect\n"
            )
        rest = ""
        # print ('repr',self.frac_coords, self.frac_coords.shape)
        for ii, i in enumerate(coords_ordered):
            rest = rest + " ".join(map(str, i)) + " " + str(props_ordered[ii]) + "\n"
        result = header + middle + rest
        return result


class Incar(object):
    """
    VASP INCAR files as python dictionary
    """

    def __init__(self, tags={}):
        self._tags = tags

    @staticmethod
    def from_file(filename="INCAR"):
        with open(filename, "r") as f:
            return Incar.from_string(f.read())

    def update(self, d={}):
        print("selftags1=", self._tags)
        if self._tags != {}:
            for i, j in self._tags.items():
                self._tags[i] = j  # .strip(' ')
        for i, j in d.items():
            self._tags[i] = j
        print("selftags2=", self._tags)
        return Incar(self._tags)

    def get(self, key="POTIM", temp=0.5):
        if key in list(self._tags.keys()):
            return self._tags[key]
        else:
            self._tags[key] = temp
            print("Setting the temp value to the key", temp, key)
            return self._tags[key]

    def to_dict(self):
        return self._tags

    def from_dict(self, data={}):
        print("data", data)
        return Incar(tags=data)

    @staticmethod
    def from_string(lines):
        text = lines.splitlines()
        tags = OrderedDict()
        for i in text:
            if "=" in i:
                tmp = i.split("=")
                tags.setdefault(tmp[0].strip(" "), tmp[1].strip(" "))

        return Incar(tags=tags)

    def __repr__(self):
        return str(self._tags)

    def write_file(self, filename):
        tags = self._tags
        lines = ""
        for i, j in tags.items():
            lines = lines + str(i) + str("=") + str(j) + "\n"
        f = open(filename, "w")
        f.write(lines)
        f.close()


class IndividualPotcarData(object):
    """
    Class for individual POTCAR file handling
    """

    def __init__(self, data={}):
        self._data = data

    def from_string(lines):
        text = lines.splitlines()
        individual_data = OrderedDict()
        individual_data["header1"] = text[0]
        individual_data["header2"] = text[2]
        individual_data["VRHFIN"] = text[3].split("=")[1]
        individual_data["element"] = text[3].split("=")[1].split(":")[0]
        individual_data["LEXCH"] = text[4].split("=")[1].replace(" ", "")
        individual_data["EATOM"] = text[5].split("=")[1].split()[0].replace(" ", "")
        individual_data["TITEL"] = text[7].split("=")[1].replace(" ", "")

        return IndividualPotcarData(data=individual_data)

    def from_file(filename="POTCAR"):
        with open(filename, "r") as f:
            return IndividualPotcarData.from_string(f.read())

    def __repr__(self):
        return pprint.pformat(self._data, indent=4)


class Potcar(object):
    def __init__(
        self,
        elements=[],
        pot_type="POT_GGA_PAW_PBE",
        pot_json_path="",
        potcar_strings={},
    ):
        self._elements = elements
        self._pot_type = pot_type
        self._potcar_strings = potcar_strings
        self._pot_json_path = pot_json_path

        if self._potcar_strings == {}:
            pot_json_file = str(
                os.path.join(os.path.dirname(__file__), "default_potcars.json")
            )
            pot_json = open(pot_json_file, "r")
            pot_json_selected = json.load(pot_json)
            pot_json.close()
            potcar_strings = OrderedDict()
            for i in self._elements:
                for j, k in pot_json_selected.items():
                    if i == j:
                        potcar_strings.setdefault(i, k)
            self._potcar_strings = potcar_strings
            if len(self._elements) != len(self._potcar_strings.keys()):
                raise ValueError(
                    "Number of elements is not same as potcar_strings",
                    self._elements,
                    self._potcar_strings.keys(),
                )
        else:
            pot_json = open(self._pot_json_path, "r")
            pot_json_selected = json.load(pot_json)
            pot_json.close()
            potcar_strings = OrderedDict()
            for i, j in pot_json_selected.items():
                if i in self._elements:
                    potcar_strings.setdefault(i, j)
            self._potcar_strings = potcar_strings
            if len(self._elements) != len(self._potcar_strings.keys()):
                raise ValueError("Number of elements is not same as potcar_strings")

    def catenate_potcar_files(self, destination_filename="POTCAR", filenames=[]):
        with open(destination_filename, "w") as outfile:
            for fname in filenames:
                with open(fname) as infile:
                    for line in infile:
                        outfile.write(line)

    def list_potcar_files(self):
        pot_files = []
        vasp_dir = str(os.environ["JARVIS_VASP_PSP_DIR"])
        vasp_psp_dir = str(os.path.join(vasp_dir, self._pot_type))
        potcar_strings = self._potcar_strings
        for i, j in potcar_strings.items():
            tmp = os.path.join(vasp_psp_dir, j, "POTCAR")
            pot_files.append(tmp)
        return pot_files

    def write_file(self, filename="POTCAR"):
        pot_files = self.list_potcar_files()
        self.catenate_potcar_files(destination_filename=filename, filenames=pot_files)

    def __repr__(self):
        return str(str(self._pot_type) + "\n" + str(self._potcar_strings))


class Kpoints(object):
    """
    VASP KPOINTS as object
    """

    def __init__(self, filename=""):
        self.filename = filename
        if filename != "":
            f = open(self.filename, "r")
            self.lines = f.read().splitlines()
            f.close()
            self.kpoints = self.read(self.lines)

    @classmethod
    def read(self, lines):

        if lines[1] == "0":
            return self.get_mesh_kp(lines=lines)
        elif "Reciprocal" in lines[2]:
            return self.get_ibz_kp(lines=lines)
        else:
            ValueError("K-point scheme is not implemented")

    def get_mesh_kp(lines=""):
        print("lines", lines)
        grid = [int(i) for i in lines[3].split()]
        # print (grid)
        kpts = generate_kgrid(grid)
        kpoints = Kpoints3D(kpoints=np.array(kpts))
        return kpoints

    def get_ibz_kp(lines=""):
        kp_labels = []
        all_kp = []
        kp_labels_points = []
        for ii, i in enumerate(lines):
            if ii > 2:
                tmp = i.split()
                k = [tmp[0], tmp[1], tmp[2]]
                all_kp.append(k)
                if len(tmp) == 5:
                    tmp = str("$") + str(tmp[4]) + str("$")
                    if len(kp_labels) == 0:
                        kp_labels.append(tmp)
                        kp_labels_points.append(ii - 3)
                    elif tmp != kp_labels[-1]:
                        kp_labels.append(tmp)
                        kp_labels_points.append(ii - 3)
        labels = []
        for i in range(len(all_kp)):
            labels.append("")
        for i, j in zip(kp_labels, kp_labels_points):
            labels[j] = i
        all_kp = np.array(all_kp, dtype="float")
        kpoints = Kpoints3D(kpoints=all_kp, labels=labels, kpoint_mode="linemode")
        return kpoints
        # print ('kp_labels',labels,len(all_kp))
        # print ('kp2',kp_labels_points, kp_labels,np.array(all_kp,dtype='float'))

        # return kp_labels_points, kp_labels,np.array(all_kp,dtype='float')


"""
if __name__ == "__main__":

    kp = open(
        "../../examples/vasp/SiOptb88/MAIN-RELAX-bulk@mp_149/KPOINTS", "r"
    )  # .read_file()
    lines = kp.read().splitlines()
    kp.close()
    print('lbl',Kpoints.read(lines).labels)
    # print (kp.kpoints)
    import sys

    sys.exit()
    kp = Kpoints(filename="../../examples/vasp/SiOptb88/MAIN-RELAX-bulk@mp_149/KPOINTS")

    p = Poscar.from_file(
        filename="/rk2/knc6/JARVIS-DFT/2D-1L/POSCAR-mp-2815-1L.vasp_PBEBO/MAIN-RELAX-Surf-mp-2815/POSCAR"
    )
    p.write_file("POSCAR")
    from jarvis.db.jsonutils import dumpjson

    dumpjson(p.atoms.to_dict(), "pp.json")
    # print (p.atoms*[3,3,3])
    # p = Incar.from_file(filename='/rk2/knc6/JARVIS-DFT/2D-1L/POSCAR-mp-2815-1L.vasp_PBEBO/MAIN-RELAX-Surf-mp-2815/INCAR')
    # print (p)
    # p.write_file('pp')
    # elements=['S','P']
    # p=Potcar(elements=elements)
    # print (p)
    # p.write_file('POTCAR')
    # inp=IndividualPotcarData.from_file('POTCAR')
    # print (inp)
"""



"""
Module for k-points used in solid-state calculations
"""

from jarvis.core.lattice import Lattice
from jarvis.core.atoms import Atoms
import numpy as np
from numpy import linalg as LA
from collections import OrderedDict
import pprint
from jarvis.analysis.structure.spacegroup import Spacegroup3D
from warnings import warn
from numpy import pi, cos, sin
from math import cos, sin, tan, e, pi, ceil
from math import ceil
#from jarvis.io.vasp.inputs import Poscar


def generate_kgrid(grid=[5, 5, 5]):
    t = []
    for i in range(grid[0]):
        for j in range(grid[1]):
            for k in range(grid[2]):
                t.append(
                    [
                        float(i) / (float(grid[0])),
                        float(j) / (float(grid[1])),
                        float(k) / (float(grid[2])),
                    ]
                )
    return t

def generate_kpath(kpath = [[0,0,0],[0,0.5,.5]],num_k=5):
        K = []
        for i in range(len(kpath)-1):
            dk = np.array(kpath[i+1]) - np.array(kpath[i])
            for j in range(num_k):
                K.append(np.array(kpath[i]) + dk * (float(j)/float(num_k)))     
        K.append(kpath[-1])
        return K

class Kpoints3D(object):
    def __init__(
        self,
        kpoints=[[1, 1, 1]],
        labels=[],
        kpoints_weights=[],
        kpoint_mode="automatic",
        header="Gamma",
    ):
        """
        Several types of k-points in the mesh or high-symmetry BZ
        """
        self._kpoints = kpoints
        self._labels = labels
        self._kpoint_mode = kpoint_mode
        self._header = header
        self._kp_weights = kpoints_weights

    def automatic_length_mesh(self, lattice_mat=[], length=20, header="Gamma"):
        inv_lat = Lattice(lattice_mat=lattice_mat).inv_lattice()
        b1 = LA.norm(np.array(inv_lat[0]))
        b2 = LA.norm(np.array(inv_lat[1]))
        b3 = LA.norm(np.array(inv_lat[2]))
        n1 = int(max(1, length * b1 + 0.5))
        n2 = int(max(1, length * b2 + 0.5))
        n3 = int(max(1, length * b3 + 0.5))
        return Kpoints3D(kpoints=[[n1, n2, n3]], header=header, kpoint_mode="automatic")

    @property
    def kpts(self):
        return self._kpoints

    @property
    def labels(self):
        return self._labels

    def write_file(self, filename=""):
        if self._kpoint_mode == "automatic":
            f = open(filename, "w")
            f.write("Automatic kpoint scheme\n")
            f.write("0\n")
            line = str(self._header) + "\n"
            f.write(line)
            kp = self._kpoints
            line = (
                str(kp[0][0])
                + str(" ")
                + str(kp[0][1])
                + str(" ")
                + str(kp[0][2])
                + str("\n")
            )
            f.write(line)
            f.close()

        if self._kpoint_mode == "linemode":
            kp = self._kpoints
            lbls = self._labels
            f = open(filename, "w")
            line = str(self._header) + "\n"
            f.write(line)
            line = str(len(kp)) + "\n"
            f.write(line)
            f.write("Reciprocal\n")
            if self._kp_weights == []:
                self._kp_weights = np.ones(len(lbls))
            for i, ii in enumerate(lbls):
                f.write(
                    "%6f %6f %6f %d %s \n"
                    % (kp[i][0], kp[i][1], kp[i][2], self._kp_weights[i], ii)
                )
            f.close()

    def to_dict(self):
        d = OrderedDict()
        d["kpoints"] = self._kpoints
        d["labels"] = self._labels
        d["kpoint_mode"] = self._kpoint_mode
        d["header"] = self._header
        return d

    def high_symm_path(self, atoms):
        spg = Spacegroup3D(atoms=atoms)
        lat_sys = spg.lattice_system
        spg_symb = spg.space_group_symbol
        kp = None
        if lat_sys == "cubic":
            if "P" in spg_symb:
                kp = HighSymmetryKpoint3DFactory().cubic()
            elif "F" in spg_symb:
                kp = HighSymmetryKpoint3DFactory().fcc()
            elif "I" in spg_symb:
                kp = HighSymmetryKpoint3DFactory().bcc()
            else:
                print("kpath space group  is not implemeted ", spg_symbol)

        elif lat_sys == "tetragonal":
            if "P" in spg_symb:
                kp = HighSymmetryKpoint3DFactory().tet()
            elif "I" in spg_symb:
                cvn = spg.conventional_standard_structure
                a = cvn.lattice.a
                c = cvn.lattice.c
                if c < a:
                    kp = HighSymmetryKpoint3DFactory().bctet1(c, a)
                else:
                    kp = HighSymmetryKpoint3DFactory().bctet2(c, a)
            else:
                print("kpath space group is not implemeted ", spg_symbol)

        elif lat_sys == "orthorhombic":
            cvn = spg.conventional_standard_structure
            a = cvn.lattice.a
            c = cvn.lattice.c
            b = cvn.lattice.b

            if "P" in spg_symb:
                kp = HighSymmetryKpoint3DFactory().orc()

            elif "F" in spg_symb:
                if 1 / a ** 2 > 1 / b ** 2 + 1 / c ** 2:
                    kp = HighSymmetryKpoint3DFactory().orcf1(a, b, c)
                elif 1 / a ** 2 < 1 / b ** 2 + 1 / c ** 2:
                    kp = HighSymmetryKpoint3DFactory().orcf2(a, b, c)
                else:
                    kp = HighSymmetryKpoint3DFactory().orcf3(a, b, c)

            elif "I" in spg_symb:
                kp = HighSymmetryKpoint3DFactory().orci(a, b, c)

            elif "C" in spg_symb or "A" in spg_symb:
                kp = HighSymmetryKpoint3DFactory().orcc(a, b, c)
            else:
                print("kpath space group is not implemeted ", spg_symbol)

        elif lat_sys == "hexagonal":
            kp = HighSymmetryKpoint3DFactory().hex()

        elif lat_sys == "rhombohedral":
            prim = spg.primitive_atoms
            alpha = prim.lattice.angles[0]
            if alpha < 90:
                kp = HighSymmetryKpoint3DFactory().rhl1(alpha * pi / 180)
            else:
                kp = HighSymmetryKpoint3DFactory().rhl2(alpha * pi / 180)

        elif lat_sys == "monoclinic":
            cvn = spg.conventional_standard_structure
            a, b, c = cvn.lattice.abc
            alpha = cvn.lattice.angles[0]

            if "P" in spg_symb:
                kp = HighSymmetryKpoint3DFactory().mcl(b, c, alpha * pi / 180)

            elif "C" in spg_symb:
                prim = spg.primitive_atoms.lattice.reciprocal_lattice()

                kgamma = prim.angles[2]
                if kgamma > 90:
                    kp = HighSymmetryKpoint3DFactory().mclc1(a, b, c, alpha * pi / 180)
                if kgamma == 90:
                    kp = HighSymmetryKpoint3DFactory().mclc2(a, b, c, alpha * pi / 180)
                if kgamma < 90:
                    if (
                        b * cos(alpha * pi / 180) / c
                        + b ** 2 * sin(alpha * pi / 180) ** 2 / a ** 2
                        < 1
                    ):
                        kp = HighSymmetryKpoint3DFactory().mclc3(
                            a, b, c, alpha * pi / 180
                        )
                    if (
                        b * cos(alpha * pi / 180) / c
                        + b ** 2 * sin(alpha * pi / 180) ** 2 / a ** 2
                        == 1
                    ):
                        kp = HighSymmetryKpoint3DFactory().mclc4(
                            a, b, c, alpha * pi / 180
                        )
                    if (
                        b * cos(alpha * pi / 180) / c
                        + b ** 2 * sin(alpha * pi / 180) ** 2 / a ** 2
                        > 1
                    ):
                        kp = HighSymmetryKpoint3DFactory().mclc5(
                            a, b, c, alpha * pi / 180
                        )
            else:
                print("kpath space group is not implemeted ", spg_symbol)

        elif lat_sys == "triclinic":
            prim = spg.primitive_atoms.lattice.reciprocal_lattice()
            kalpha = prim.angles[0]
            kbeta = prim.angles[1]
            kgamma = prim.angles[2]
            if kalpha > 90 and kbeta > 90 and kgamma > 90:
                kp = HighSymmetryKpoint3DFactory().tria()
            if kalpha < 90 and kbeta < 90 and kgamma < 90:
                kp = HighSymmetryKpoint3DFactory().trib()
            if kalpha > 90 and kbeta > 90 and kgamma == 90:
                kp = HighSymmetryKpoint3DFactory().tria()
            if kalpha < 90 and kbeta < 90 and kgamma == 90:
                kp = HighSymmetryKpoint3DFactory().trib()

        else:
            print("kpath space group is not implemeted ", spg_symbol)
        # print("kp",spg_symb)
        return kp

    def high_kpath(self, atoms):
        return self.high_symm_path(atoms).as_dict()

    def interpolated_points(self, atoms, line_density=20, coords_are_cartesian=False):
        """
        Useful for bandstructure, controlled by the line_density
        """
        list_k_points = []
        sym_point_labels = []
        spg = Spacegroup3D(atoms=atoms)
        prim = spg.primitive_atoms
        self.kpath = self.high_kpath(atoms)
        self._prim_rec = prim.lattice.reciprocal_lattice()
        # print ('self._prim_rec',self._prim_rec)
        for b in self.kpath["path"]:
            for i in range(1, len(b)):
                start = np.array(self.kpath["kpoints"][b[i - 1]])
                end = np.array(self.kpath["kpoints"][b[i]])
                distance = np.linalg.norm(
                    self._prim_rec.cart_coords(start) - self._prim_rec.cart_coords(end)
                )
                nb = int(ceil(distance * line_density))
                sym_point_labels.extend([b[i - 1]] + [""] * (nb - 1) + [b[i]])
                list_k_points.extend(
                    [
                        self._prim_rec.cart_coords(start)
                        + float(i)
                        / float(nb)
                        * (
                            self._prim_rec.cart_coords(end)
                            - self._prim_rec.cart_coords(start)
                        )
                        for i in range(0, nb + 1)
                    ]
                )
        if coords_are_cartesian:
            return list_k_points, sym_point_labels
        else:
            frac_k_points = [self._prim_rec.frac_coords(k) for k in list_k_points]
            return frac_k_points, sym_point_labels

    def kpath(
        self,
        atoms,
        line_density=20,
        weights=[],
        unique_kp_only=False,
        coords_are_cartesian=False,
    ):
        k_points, labels = self.interpolated_points(
            atoms, line_density=line_density, coords_are_cartesian=coords_are_cartesian
        )
        if unique_kp_only:
            uniqueValues, indicesList = np.unique(k_points, axis=0, return_index=True)
            k_points = np.array(k_points)[indicesList.astype(int)]
            labels = np.array(labels)[indicesList.astype(int)]
        return Kpoints3D(
            kpoints=k_points,
            header="Non SCF run along symmetry lines",
            labels=labels,
            kpoint_mode="linemode",
        )

    def __repr__(self, indent=4):
        return pprint.pformat(self.to_dict(), indent=indent)


class HighSymmetryKpoint3DFactory(object):
    def __init__(self, kpoints=[], path=[]):
        """
        High-symmetry k-points for different crystal-systems, similar to AFLOW and Pymatgen
        """
        self._kpoints = kpoints
        self._path = path

    def cubic(self):
        """
        Cubic HighSymmKPath
        :return: Dict
        """
        self.name = "CUB"
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "X": np.array([0.0, 0.5, 0.0]),
            "R": np.array([0.5, 0.5, 0.5]),
            "M": np.array([0.5, 0.5, 0.0]),
        }
        path = [["\\Gamma", "X", "M", "\\Gamma", "R", "X"], ["M", "R"]]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def fcc(self):
        """
        Fcc HighSymmKPath
        :return: Dict
        """
        self.name = "FCC"
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "K": np.array([3.0 / 8.0, 3.0 / 8.0, 3.0 / 4.0]),
            "L": np.array([0.5, 0.5, 0.5]),
            "U": np.array([5.0 / 8.0, 1.0 / 4.0, 5.0 / 8.0]),
            "W": np.array([0.5, 1.0 / 4.0, 3.0 / 4.0]),
            "X": np.array([0.5, 0.0, 0.5]),
        }
        path = [
            ["\\Gamma", "X", "W", "K", "\\Gamma", "L", "U", "W", "L", "K"],
            ["U", "X"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def bcc(self):
        """
        Bcc HighSymmKPath
        :return: Dict
        """
        self.name = "BCC"
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "H": np.array([0.5, -0.5, 0.5]),
            "P": np.array([0.25, 0.25, 0.25]),
            "N": np.array([0.0, 0.0, 0.5]),
        }
        path = [["\\Gamma", "H", "N", "\\Gamma", "P", "H"], ["P", "N"]]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def as_dict(self):
        d = OrderedDict()
        d["kpoints"] = self._kpoints
        d["path"] = self._path
        return d

    def tet(self):
        """
        Tetragonal HighSymmKPath
        :return: Dict
        """
        self.name = "TET"
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "A": np.array([0.5, 0.5, 0.5]),
            "M": np.array([0.5, 0.5, 0.0]),
            "R": np.array([0.0, 0.5, 0.5]),
            "X": np.array([0.0, 0.5, 0.0]),
            "Z": np.array([0.0, 0.0, 0.5]),
        }
        path = [
            ["\\Gamma", "X", "M", "\\Gamma", "Z", "R", "A", "Z"],
            ["X", "R"],
            ["M", "A"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def bctet1(self, c, a):
        """
        BCT1 HighSymmKPath
        :return: Dict
        """
        self.name = "BCT1"
        eta = (1 + c ** 2 / a ** 2) / 4.0
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "M": np.array([-0.5, 0.5, 0.5]),
            "N": np.array([0.0, 0.5, 0.0]),
            "P": np.array([0.25, 0.25, 0.25]),
            "X": np.array([0.0, 0.0, 0.5]),
            "Z": np.array([eta, eta, -eta]),
            "Z_1": np.array([-eta, 1 - eta, eta]),
        }
        path = [["\\Gamma", "X", "M", "\\Gamma", "Z", "P", "N", "Z_1", "M"], ["X", "P"]]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def bctet2(self, c, a):
        """
        BCT2 HighSymmKPath
        :return: Dict
        """
        self.name = "BCT2"
        eta = (1 + a ** 2 / c ** 2) / 4.0
        zeta = a ** 2 / (2 * c ** 2)
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "N": np.array([0.0, 0.5, 0.0]),
            "P": np.array([0.25, 0.25, 0.25]),
            "\\Sigma": np.array([-eta, eta, eta]),
            "\\Sigma_1": np.array([eta, 1 - eta, -eta]),
            "X": np.array([0.0, 0.0, 0.5]),
            "Y": np.array([-zeta, zeta, 0.5]),
            "Y_1": np.array([0.5, 0.5, -zeta]),
            "Z": np.array([0.5, 0.5, -0.5]),
        }
        path = [
            [
                "\\Gamma",
                "X",
                "Y",
                "\\Sigma",
                "\\Gamma",
                "Z",
                "\\Sigma_1",
                "N",
                "P",
                "Y_1",
                "Z",
            ],
            ["X", "P"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def orc(self):
        """
        Orthorhombic HighSymmKPath
        :return: Dict
        """
        self.name = "ORC"
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "R": np.array([0.5, 0.5, 0.5]),
            "S": np.array([0.5, 0.5, 0.0]),
            "T": np.array([0.0, 0.5, 0.5]),
            "U": np.array([0.5, 0.0, 0.5]),
            "X": np.array([0.5, 0.0, 0.0]),
            "Y": np.array([0.0, 0.5, 0.0]),
            "Z": np.array([0.0, 0.0, 0.5]),
        }
        path = [
            ["\\Gamma", "X", "S", "Y", "\\Gamma", "Z", "U", "R", "T", "Z"],
            ["Y", "T"],
            ["U", "X"],
            ["S", "R"],
        ]

        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def orcf1(self, a, b, c):
        """
        Orthorhombic f1 HighSymmKPath
        :return: Dict
        """
        self.name = "ORCF1"
        zeta = (1 + a ** 2 / b ** 2 - a ** 2 / c ** 2) / 4
        eta = (1 + a ** 2 / b ** 2 + a ** 2 / c ** 2) / 4

        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "A": np.array([0.5, 0.5 + zeta, zeta]),
            "A_1": np.array([0.5, 0.5 - zeta, 1 - zeta]),
            "L": np.array([0.5, 0.5, 0.5]),
            "T": np.array([1, 0.5, 0.5]),
            "X": np.array([0.0, eta, eta]),
            "X_1": np.array([1, 1 - eta, 1 - eta]),
            "Y": np.array([0.5, 0.0, 0.5]),
            "Z": np.array([0.5, 0.5, 0.0]),
        }
        path = [
            ["\\Gamma", "Y", "T", "Z", "\\Gamma", "X", "A_1", "Y"],
            ["T", "X_1"],
            ["X", "A", "Z"],
            ["L", "\\Gamma"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def orcf2(self, a, b, c):
        """
        Orthorhombic f2 HighSymmKPath
        :return: Dict
        """
        self.name = "ORCF2"
        phi = (1 + c ** 2 / b ** 2 - c ** 2 / a ** 2) / 4
        eta = (1 + a ** 2 / b ** 2 - a ** 2 / c ** 2) / 4
        delta = (1 + b ** 2 / a ** 2 - b ** 2 / c ** 2) / 4
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "C": np.array([0.5, 0.5 - eta, 1 - eta]),
            "C_1": np.array([0.5, 0.5 + eta, eta]),
            "D": np.array([0.5 - delta, 0.5, 1 - delta]),
            "D_1": np.array([0.5 + delta, 0.5, delta]),
            "L": np.array([0.5, 0.5, 0.5]),
            "H": np.array([1 - phi, 0.5 - phi, 0.5]),
            "H_1": np.array([phi, 0.5 + phi, 0.5]),
            "X": np.array([0.0, 0.5, 0.5]),
            "Y": np.array([0.5, 0.0, 0.5]),
            "Z": np.array([0.5, 0.5, 0.0]),
        }
        path = [
            ["\\Gamma", "Y", "C", "D", "X", "\\Gamma", "Z", "D_1", "H", "C"],
            ["C_1", "Z"],
            ["X", "H_1"],
            ["H", "Y"],
            ["L", "\\Gamma"],
        ]

        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def orcf3(self, a, b, c):
        """
        Orthorhombic f3 HighSymmKPath
        :return: Dict
        """
        self.name = "ORCF3"
        zeta = (1 + a ** 2 / b ** 2 - a ** 2 / c ** 2) / 4
        eta = (1 + a ** 2 / b ** 2 + a ** 2 / c ** 2) / 4
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "A": np.array([0.5, 0.5 + zeta, zeta]),
            "A_1": np.array([0.5, 0.5 - zeta, 1 - zeta]),
            "L": np.array([0.5, 0.5, 0.5]),
            "T": np.array([1, 0.5, 0.5]),
            "X": np.array([0.0, eta, eta]),
            "X_1": np.array([1, 1 - eta, 1 - eta]),
            "Y": np.array([0.5, 0.0, 0.5]),
            "Z": np.array([0.5, 0.5, 0.0]),
        }
        path = [
            ["\\Gamma", "Y", "T", "Z", "\\Gamma", "X", "A_1", "Y"],
            ["X", "A", "Z"],
            ["L", "\\Gamma"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def orci(self, a, b, c):
        """
        Orthorhombic I HighSymmKPath
        :return: Dict
        """
        self.name = "ORCI"
        zeta = (1 + a ** 2 / c ** 2) / 4
        eta = (1 + b ** 2 / c ** 2) / 4
        delta = (b ** 2 - a ** 2) / (4 * c ** 2)
        mu = (a ** 2 + b ** 2) / (4 * c ** 2)
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "L": np.array([-mu, mu, 0.5 - delta]),
            "L_1": np.array([mu, -mu, 0.5 + delta]),
            "L_2": np.array([0.5 - delta, 0.5 + delta, -mu]),
            "R": np.array([0.0, 0.5, 0.0]),
            "S": np.array([0.5, 0.0, 0.0]),
            "T": np.array([0.0, 0.0, 0.5]),
            "W": np.array([0.25, 0.25, 0.25]),
            "X": np.array([-zeta, zeta, zeta]),
            "X_1": np.array([zeta, 1 - zeta, -zeta]),
            "Y": np.array([eta, -eta, eta]),
            "Y_1": np.array([1 - eta, eta, -eta]),
            "Z": np.array([0.5, 0.5, -0.5]),
        }
        path = [
            ["\\Gamma", "X", "L", "T", "W", "R", "X_1", "Z", "\\Gamma", "Y", "S", "W"],
            ["L_1", "Y"],
            ["Y_1", "Z"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def orcc(self, a, b, c):
        """
        Orthorhombic C HighSymmKPath
        :return: Dict
        """
        self.name = "ORCC"
        zeta = (1 + a ** 2 / b ** 2) / 4
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "A": np.array([zeta, zeta, 0.5]),
            "A_1": np.array([-zeta, 1 - zeta, 0.5]),
            "R": np.array([0.0, 0.5, 0.5]),
            "S": np.array([0.0, 0.5, 0.0]),
            "T": np.array([-0.5, 0.5, 0.5]),
            "X": np.array([zeta, zeta, 0.0]),
            "X_1": np.array([-zeta, 1 - zeta, 0.0]),
            "Y": np.array([-0.5, 0.5, 0]),
            "Z": np.array([0.0, 0.0, 0.5]),
        }
        path = [
            [
                "\\Gamma",
                "X",
                "S",
                "R",
                "A",
                "Z",
                "\\Gamma",
                "Y",
                "X_1",
                "A_1",
                "T",
                "Y",
            ],
            ["Z", "T"],
        ]

        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def hex(self):
        """
        Hexagonal HighSymmKPath
        :return: Dict
        """
        self.name = "HEX"
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "A": np.array([0.0, 0.0, 0.5]),
            "H": np.array([1.0 / 3.0, 1.0 / 3.0, 0.5]),
            "K": np.array([1.0 / 3.0, 1.0 / 3.0, 0.0]),
            "L": np.array([0.5, 0.0, 0.5]),
            "M": np.array([0.5, 0.0, 0.0]),
        }
        path = [
            ["\\Gamma", "M", "K", "\\Gamma", "A", "L", "H", "A"],
            ["L", "M"],
            ["K", "H"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def rhl1(self, alpha):
        """
        Rhombohedral 1 HighSymmKPath
        :return: Dict
        """
        self.name = "RHL1"
        eta = (1 + 4 * cos(alpha)) / (2 + 4 * cos(alpha))
        nu = 3.0 / 4.0 - eta / 2.0
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "B": np.array([eta, 0.5, 1.0 - eta]),
            "B_1": np.array([1.0 / 2.0, 1.0 - eta, eta - 1.0]),
            "F": np.array([0.5, 0.5, 0.0]),
            "L": np.array([0.5, 0.0, 0.0]),
            "L_1": np.array([0.0, 0.0, -0.5]),
            "P": np.array([eta, nu, nu]),
            "P_1": np.array([1.0 - nu, 1.0 - nu, 1.0 - eta]),
            "P_2": np.array([nu, nu, eta - 1.0]),
            "Q": np.array([1.0 - nu, nu, 0.0]),
            "X": np.array([nu, 0.0, -nu]),
            "Z": np.array([0.5, 0.5, 0.5]),
        }
        path = [
            ["\\Gamma", "L", "B_1"],
            ["B", "Z", "\\Gamma", "X"],
            ["Q", "F", "P_1", "Z"],
            ["L", "P"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def rhl2(self, alpha):
        """
        Rhombohedral 2 HighSymmKPath
        :return: Dict
        """
        self.name = "RHL2"
        eta = 1 / (2 * tan(alpha / 2.0) ** 2)
        nu = 3.0 / 4.0 - eta / 2.0
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "F": np.array([0.5, -0.5, 0.0]),
            "L": np.array([0.5, 0.0, 0.0]),
            "P": np.array([1 - nu, -nu, 1 - nu]),
            "P_1": np.array([nu, nu - 1.0, nu - 1.0]),
            "Q": np.array([eta, eta, eta]),
            "Q_1": np.array([1.0 - eta, -eta, -eta]),
            "Z": np.array([0.5, -0.5, 0.5]),
        }
        path = [["\\Gamma", "P", "Z", "Q", "\\Gamma", "F", "P_1", "Q_1", "L", "Z"]]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def mcl(self, b, c, beta):
        """
        Monoclinic 1 HighSymmKPath
        :return: Dict
        """
        self.name = "MCL"
        eta = (1 - b * cos(beta) / c) / (2 * sin(beta) ** 2)
        nu = 0.5 - eta * c * cos(beta) / b
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "A": np.array([0.5, 0.5, 0.0]),
            "C": np.array([0.0, 0.5, 0.5]),
            "D": np.array([0.5, 0.0, 0.5]),
            "D_1": np.array([0.5, 0.5, -0.5]),
            "E": np.array([0.5, 0.5, 0.5]),
            "H": np.array([0.0, eta, 1.0 - nu]),
            "H_1": np.array([0.0, 1.0 - eta, nu]),
            "H_2": np.array([0.0, eta, -nu]),
            "M": np.array([0.5, eta, 1.0 - nu]),
            "M_1": np.array([0.5, 1 - eta, nu]),
            "M_2": np.array([0.5, 1 - eta, nu]),
            "X": np.array([0.0, 0.5, 0.0]),
            "Y": np.array([0.0, 0.0, 0.5]),
            "Y_1": np.array([0.0, 0.0, -0.5]),
            "Z": np.array([0.5, 0.0, 0.0]),
        }
        path = [
            ["\\Gamma", "Y", "H", "C", "E", "M_1", "A", "X", "H_1"],
            ["M", "D", "Z"],
            ["Y", "D"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def mclc1(self, a, b, c, alpha):
        """
        Monoclinic C1 HighSymmKPath
        :return: Dict
        """
        self.name = "MCLC1"
        zeta = (2 - b * cos(alpha) / c) / (4 * sin(alpha) ** 2)
        eta = 0.5 + 2 * zeta * c * cos(alpha) / b
        psi = 0.75 - a ** 2 / (4 * b ** 2 * sin(alpha) ** 2)
        phi = psi + (0.75 - psi) * b * cos(alpha) / c
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "N": np.array([0.5, 0.0, 0.0]),
            "N_1": np.array([0.0, -0.5, 0.0]),
            "F": np.array([1 - zeta, 1 - zeta, 1 - eta]),
            "F_1": np.array([zeta, zeta, eta]),
            "F_2": np.array([-zeta, -zeta, 1 - eta]),
            # 'F_3': np.array([1 - zeta, -zeta, 1 - eta]),
            "I": np.array([phi, 1 - phi, 0.5]),
            "I_1": np.array([1 - phi, phi - 1, 0.5]),
            "L": np.array([0.5, 0.5, 0.5]),
            "M": np.array([0.5, 0.0, 0.5]),
            "X": np.array([1 - psi, psi - 1, 0.0]),
            "X_1": np.array([psi, 1 - psi, 0.0]),
            "X_2": np.array([psi - 1, -psi, 0.0]),
            "Y": np.array([0.5, 0.5, 0.0]),
            "Y_1": np.array([-0.5, -0.5, 0.0]),
            "Z": np.array([0.0, 0.0, 0.5]),
        }
        path = [
            ["\\Gamma", "Y", "F", "L", "I"],
            ["I_1", "Z", "F_1"],
            ["Y", "X_1"],
            ["X", "\\Gamma", "N"],
            ["M", "\\Gamma"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def mclc2(self, a, b, c, alpha):
        """
        Monoclinic C2 HighSymmKPath
        :return: Dict
        """
        self.name = "MCLC2"
        zeta = (2 - b * cos(alpha) / c) / (4 * sin(alpha) ** 2)
        eta = 0.5 + 2 * zeta * c * cos(alpha) / b
        psi = 0.75 - a ** 2 / (4 * b ** 2 * sin(alpha) ** 2)
        phi = psi + (0.75 - psi) * b * cos(alpha) / c
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "N": np.array([0.5, 0.0, 0.0]),
            "N_1": np.array([0.0, -0.5, 0.0]),
            "F": np.array([1 - zeta, 1 - zeta, 1 - eta]),
            "F_1": np.array([zeta, zeta, eta]),
            "F_2": np.array([-zeta, -zeta, 1 - eta]),
            "F_3": np.array([1 - zeta, -zeta, 1 - eta]),
            "I": np.array([phi, 1 - phi, 0.5]),
            "I_1": np.array([1 - phi, phi - 1, 0.5]),
            "L": np.array([0.5, 0.5, 0.5]),
            "M": np.array([0.5, 0.0, 0.5]),
            "X": np.array([1 - psi, psi - 1, 0.0]),
            "X_1": np.array([psi, 1 - psi, 0.0]),
            "X_2": np.array([psi - 1, -psi, 0.0]),
            "Y": np.array([0.5, 0.5, 0.0]),
            "Y_1": np.array([-0.5, -0.5, 0.0]),
            "Z": np.array([0.0, 0.0, 0.5]),
        }
        path = [
            ["\\Gamma", "Y", "F", "L", "I"],
            ["I_1", "Z", "F_1"],
            ["N", "\\Gamma", "M"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def mclc3(self, a, b, c, alpha):
        """
        Monoclinic C3 HighSymmKPath
        :return: Dict
        """
        self.name = "MCLC3"
        mu = (1 + b ** 2 / a ** 2) / 4.0
        delta = b * c * cos(alpha) / (2 * a ** 2)
        zeta = mu - 0.25 + (1 - b * cos(alpha) / c) / (4 * sin(alpha) ** 2)
        eta = 0.5 + 2 * zeta * c * cos(alpha) / b
        phi = 1 + zeta - 2 * mu
        psi = eta - 2 * delta
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "F": np.array([1 - phi, 1 - phi, 1 - psi]),
            "F_1": np.array([phi, phi - 1, psi]),
            "F_2": np.array([1 - phi, -phi, 1 - psi]),
            "H": np.array([zeta, zeta, eta]),
            "H_1": np.array([1 - zeta, -zeta, 1 - eta]),
            "H_2": np.array([-zeta, -zeta, 1 - eta]),
            "I": np.array([0.5, -0.5, 0.5]),
            "M": np.array([0.5, 0.0, 0.5]),
            "N": np.array([0.5, 0.0, 0.0]),
            "N_1": np.array([0.0, -0.5, 0.0]),
            "X": np.array([0.5, -0.5, 0.0]),
            "Y": np.array([mu, mu, delta]),
            "Y_1": np.array([1 - mu, -mu, -delta]),
            "Y_2": np.array([-mu, -mu, -delta]),
            "Y_3": np.array([mu, mu - 1, delta]),
            "Z": np.array([0.0, 0.0, 0.5]),
        }
        path = [
            ["\\Gamma", "Y", "F", "H", "Z", "I", "F_1"],
            ["H_1", "Y_1", "X", "\\Gamma", "N"],
            ["M", "\\Gamma"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def mclc4(self, a, b, c, alpha):
        """
        Monoclinic C4 HighSymmKPath
        :return: Dict
        """
        self.name = "MCLC4"
        mu = (1 + b ** 2 / a ** 2) / 4.0
        delta = b * c * cos(alpha) / (2 * a ** 2)
        zeta = mu - 0.25 + (1 - b * cos(alpha) / c) / (4 * sin(alpha) ** 2)
        eta = 0.5 + 2 * zeta * c * cos(alpha) / b
        phi = 1 + zeta - 2 * mu
        psi = eta - 2 * delta
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "F": np.array([1 - phi, 1 - phi, 1 - psi]),
            "F_1": np.array([phi, phi - 1, psi]),
            "F_2": np.array([1 - phi, -phi, 1 - psi]),
            "H": np.array([zeta, zeta, eta]),
            "H_1": np.array([1 - zeta, -zeta, 1 - eta]),
            "H_2": np.array([-zeta, -zeta, 1 - eta]),
            "I": np.array([0.5, -0.5, 0.5]),
            "M": np.array([0.5, 0.0, 0.5]),
            "N": np.array([0.5, 0.0, 0.0]),
            "N_1": np.array([0.0, -0.5, 0.0]),
            "X": np.array([0.5, -0.5, 0.0]),
            "Y": np.array([mu, mu, delta]),
            "Y_1": np.array([1 - mu, -mu, -delta]),
            "Y_2": np.array([-mu, -mu, -delta]),
            "Y_3": np.array([mu, mu - 1, delta]),
            "Z": np.array([0.0, 0.0, 0.5]),
        }
        path = [
            ["\\Gamma", "Y", "F", "H", "Z", "I"],
            ["H_1", "Y_1", "X", "\\Gamma", "N"],
            ["M", "\\Gamma"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def mclc5(self, a, b, c, alpha):
        """
        Monoclinic C5 HighSymmKPath
        :return: Dict
        """
        self.name = "MCLC5"
        zeta = (b ** 2 / a ** 2 + (1 - b * cos(alpha) / c) / sin(alpha) ** 2) / 4
        eta = 0.5 + 2 * zeta * c * cos(alpha) / b
        mu = eta / 2 + b ** 2 / (4 * a ** 2) - b * c * cos(alpha) / (2 * a ** 2)
        nu = 2 * mu - zeta
        rho = 1 - zeta * a ** 2 / b ** 2
        omega = (
            (4 * nu - 1 - b ** 2 * sin(alpha) ** 2 / a ** 2) * c / (2 * b * cos(alpha))
        )
        delta = zeta * c * cos(alpha) / b + omega / 2 - 0.25
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "F": np.array([nu, nu, omega]),
            "F_1": np.array([1 - nu, 1 - nu, 1 - omega]),
            "F_2": np.array([nu, nu - 1, omega]),
            "H": np.array([zeta, zeta, eta]),
            "H_1": np.array([1 - zeta, -zeta, 1 - eta]),
            "H_2": np.array([-zeta, -zeta, 1 - eta]),
            "I": np.array([rho, 1 - rho, 0.5]),
            "I_1": np.array([1 - rho, rho - 1, 0.5]),
            "L": np.array([0.5, 0.5, 0.5]),
            "M": np.array([0.5, 0.0, 0.5]),
            "N": np.array([0.5, 0.0, 0.0]),
            "N_1": np.array([0.0, -0.5, 0.0]),
            "X": np.array([0.5, -0.5, 0.0]),
            "Y": np.array([mu, mu, delta]),
            "Y_1": np.array([1 - mu, -mu, -delta]),
            "Y_2": np.array([-mu, -mu, -delta]),
            "Y_3": np.array([mu, mu - 1, delta]),
            "Z": np.array([0.0, 0.0, 0.5]),
        }
        path = [
            ["\\Gamma", "Y", "F", "L", "I"],
            ["I_1", "Z", "H", "F_1"],
            ["H_1", "Y_1", "X", "\\Gamma", "N"],
            ["M", "\\Gamma"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def tria(self):
        """
        Trigonal a HighSymmKPath
        :return: Dict
        """
        self.name = "TRI1a"
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "L": np.array([0.5, 0.5, 0.0]),
            "M": np.array([0.0, 0.5, 0.5]),
            "N": np.array([0.5, 0.0, 0.5]),
            "R": np.array([0.5, 0.5, 0.5]),
            "X": np.array([0.5, 0.0, 0.0]),
            "Y": np.array([0.0, 0.5, 0.0]),
            "Z": np.array([0.0, 0.0, 0.5]),
        }
        path = [
            ["X", "\\Gamma", "Y"],
            ["L", "\\Gamma", "Z"],
            ["N", "\\Gamma", "M"],
            ["R", "\\Gamma"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)

    def trib(self):
        """
        Trigonal b HighSymmKPath
        :return: Dict
        """
        self.name = "TRI1b"
        kpoints = {
            "\\Gamma": np.array([0.0, 0.0, 0.0]),
            "L": np.array([0.5, -0.5, 0.0]),
            "M": np.array([0.0, 0.0, 0.5]),
            "N": np.array([-0.5, -0.5, 0.5]),
            "R": np.array([0.0, -0.5, 0.5]),
            "X": np.array([0.0, -0.5, 0.0]),
            "Y": np.array([0.5, 0.0, 0.0]),
            "Z": np.array([-0.5, 0.0, 0.5]),
        }
        path = [
            ["X", "\\Gamma", "Y"],
            ["L", "\\Gamma", "Z"],
            ["N", "\\Gamma", "M"],
            ["R", "\\Gamma"],
        ]
        return HighSymmetryKpoint3DFactory(kpoints=kpoints, path=path)


"""
if __name__ == "__main__":
    box = [[2.715, 2.715, 0], [0, 2.715, 2.715], [2.715, 0, 2.715]]
    coords = [[0, 0, 0], [0.25, 0.25, 0.25]]
    elements = ["Si", "Si"]
    Si = Atoms(lattice_mat=box, coords=coords, elements=elements)
    lattice_mat = Si.lattice_mat
    kp = Kpoints3D().automatic_length_mesh(lattice_mat=lattice_mat)
    # print(kp.__repr__(0))
    kp.write_file("KPOINTS")
    sym = kp.high_symm_path(Si)._path
    # print (sym)
    x, y = kp.interpolated_points(Si)
    # for i,j in zip(x,y):
    #   print (i,j)
    Si = Poscar.from_file(
        "/rk2/knc6/JARVIS-DFT/TE-bulk/mp-541837_bulk_PBEBO/MAIN-RELAX-bulk@mp_541837/CONTCAR"
    ).atoms
    kp = Kpoints3D().kpath(atoms=Si)  # automatic_length_mesh(lattice_mat=lattice_mat)
    kp.write_file("KPOINTS")
"""



"""
Downloads files from https://figshare.com/authors/Kamal_Choudhary/4445539
"""

import zipfile, os, requests
from jarvis.db.jsonutils import loadjson

def datasets(dataset=''):
    if dataset == 'dft_2d':
        url = "https://ndownloader.figshare.com/files/22471019"
        js_tag = 'jdft_2d-4-26-2020.json'
        print ('Downloading 2D dataset ...')
    elif dataset == 'dft_3d':
        url = "https://ndownloader.figshare.com/files/22471022"
        js_tag = 'jdft_3d-4-26-2020.json' 
        print ('Downloading 3D dataset ...')

    elif dataset == 'cfid_3d':
        url = "https://ndownloader.figshare.com/files/22470818"
        js_tag = 'jml_3d-4-26-2020.json' 
        print ('Downloading 3D CFID dataset ...')
    else:
        ValueError('Dataset doesnt exist',dataset)
    return url, js_tag

    

def data(dataset='dft_2d') :
    url, js_tag = datasets(dataset)
    path = str(os.path.join(os.path.dirname(__file__),js_tag ))
    if not  os.path.isfile(path):
        zfile = str(os.path.join(os.path.dirname(__file__), "tmp.zip"))
        r = requests.get(url)
        f = open(zfile, "wb")
        f.write(r.content)
        f.close()

        with zipfile.ZipFile(zfile, 'r') as zipObj:
            #zipObj.extract(path)
            zipObj.extractall(os.path.join(os.path.dirname(__file__)))
        os.remove(zfile)
    data = loadjson(path)
    return data




def get_ff_eneleast():
    jff1 = str(os.path.join(os.path.dirname(__file__), "jff1.json"))
    if not os.path.isfile(jff1):
        r = requests.get("https://ndownloader.figshare.com/files/10307139")
        f = open(jff1, "wb")
        f.write(r.content)
        f.close()
    data_ff1 = loadjson(jff1)
    return data_ff1




if __name__ == "__main__":

    data_2d = data(dataset='dft_2d')
    print('2d',len(data_2d))
    data_3d = data(dataset='dft_3d')
    print('3d',len(data_3d))
    data_ml = data(dataset='cfid_3d')
    print('cfid3d',len(data_ml))
    data_ff = get_ff_eneleast()
    print ('ff',len(data_ff))
